arXiv is now an independent nonprofit! Learn more
License: arXiv.org perpetual non-exclusive license
arXiv:2205.10046v1 [physics.comp-ph] 20 May 2022

GPUMD: A package for constructing accurate machine-learned potentials and performing highly efficient atomistic simulations

Zheyong Fan Email: brucenju@gmail.com Affiliation: College of Physical Science and Technology, Bohai University, Jinzhou 121013, P. R. China    Yanzhou Wang Affiliation: MSP group, QTF Centre of Excellence, Department of Applied Physics, Aalto University, FI-00076 Aalto, Espoo, Finland Affiliation: Beijing Advanced Innovation Center for Materials Genome Engineering, University of Science and Technology Beijing, Beijing 100083, China    Penghua Ying Affiliation: School of Science, Harbin Institute of Technology, Shenzhen, 518055, P. R. China    Keke Song Affiliation: Beijing Advanced Innovation Center for Materials Genome Engineering, University of Science and Technology Beijing, Beijing 100083, China    Junjie Wang Affiliation: National Laboratory of Solid State Microstructures, School of Physics and Collaborative Innovation Center of Advanced Microstructures, Nanjing University, Nanjing 210093, China    Yong Wang Affiliation: National Laboratory of Solid State Microstructures, School of Physics and Collaborative Innovation Center of Advanced Microstructures, Nanjing University, Nanjing 210093, China    Zezhu Zeng Affiliation: Department of Mechanical Engineering, The University of Hong Kong, Pokfulam Road, Hong Kong SAR, China    Ke Xu Affiliation: Department of Physics, Research Institute for Biomimetics and Soft Matter, Jiujiang Research Institute and Fujian Provincial Key Laboratory for Soft Functional Materials Research, Xiamen University, Xiamen 361005, PR China.    Eric Lindgren Affiliation:  Chalmers University of Technology, Department of Physics, 41926 Gothenburg, Sweden    J. Magnus Rahm Affiliation:  Chalmers University of Technology, Department of Physics, 41926 Gothenburg, Sweden    Alexander J. Gabourie Affiliation: Department of Electrical Engineering, Stanford University, Stanford, California 94305, USA    Jiahui Liu Affiliation: Beijing Advanced Innovation Center for Materials Genome Engineering, University of Science and Technology Beijing, Beijing 100083, China    Haikuan Dong Affiliation: College of Physical Science and Technology, Bohai University, Jinzhou 121013, P. R. China Affiliation: Beijing Advanced Innovation Center for Materials Genome Engineering, University of Science and Technology Beijing, Beijing 100083, China    Jianyang Wu Affiliation: Department of Physics, Research Institute for Biomimetics and Soft Matter, Jiujiang Research Institute and Fujian Provincial Key Laboratory for Soft Functional Materials Research, Xiamen University, Xiamen 361005, PR China.    Yue Chen Affiliation: Department of Mechanical Engineering, The University of Hong Kong, Pokfulam Road, Hong Kong SAR, China    Zheng Zhong Affiliation: School of Science, Harbin Institute of Technology, Shenzhen, 518055, P. R. China    Jian Sun Email: jiansun@nju.edu.cn Affiliation: National Laboratory of Solid State Microstructures, School of Physics and Collaborative Innovation Center of Advanced Microstructures, Nanjing University, Nanjing 210093, China    Paul Erhart Email: erhart@chalmers.se Affiliation:  Chalmers University of Technology, Department of Physics, 41926 Gothenburg, Sweden    Yanjing Su Email: yjsu@ustb.edu.cn Affiliation: Beijing Advanced Innovation Center for Materials Genome Engineering, University of Science and Technology Beijing, Beijing 100083, China    Tapio Ala-Nissila Affiliation: MSP group, QTF Centre of Excellence, Department of Applied Physics, Aalto University, FI-00076 Aalto, Espoo, Finland Affiliation: Interdisciplinary Centre for Mathematical Modelling, Department of Mathematical Sciences, Loughborough University, Loughborough, Leicestershire LE11 3TU, UK
August 24, 2026
Abstract

We present our latest advancements of machine-learned potentials (MLPs) based on the neuroevolution potential (NEP) framework introduced in [Fan et al., Phys. Rev. B 104, 104309 (2021)] and their implementation in the open-source package gpumd. We increase the accuracy of NEP models both by improving the radial functions in the atomic-environment descriptor using a linear combination of Chebyshev basis functions and by extending the angular descriptor with some four-body and five-body contributions as in the atomic cluster expansion approach. We also detail our efficient implementation of the NEP approach in graphics processing units as well as our workflow for the construction of NEP models, and we demonstrate their application in large-scale atomistic simulations. By comparing to state-of-the-art MLPs, we show that the NEP approach not only achieves above-average accuracy but also is far more computationally efficient. These results demonstrate that the gpumd package is a promising tool for solving challenging problems requiring highly accurate, large-scale atomistic simulations. To enable the construction of MLPs using a minimal training set, we propose an active-learning scheme based on the latent space of a pre-trained NEP model. Finally, we introduce two separate Python packages, gpyumd and calorine, which enable the integration of gpumd into Python workflows.

I Introduction

Machine-learned classical potentials [1, 2, 3, 4, 5, 6, 7] have shown great promise in enabling accurate atomistic simulations far beyond the space and time scales that can be achieved using quantum mechanical calculations. Many open-source computer packages for machine-learned potentials (MLPs) have been published, including quip-gap [8, 9, 10], snap [11], amp [12], aenet[13, 14], ani [15], SchNet [16], DeepMD-kit [17, 18, 19], TensorMol[20], PhysNet [21], MEGNet [22], turboGAP [23], sGDML [24], n2p2 [25], simple-nn [26], panna [27], fchl [28], PiNN [29], mlip [30, 31], reann [32], tabGAP [33], pyXtal-FF [34], python-ace [35, 36], and kliff [37]. However, most existing implementations of MLPs (tabGAP [33] is a notable exception) have a computational speed of about 1×103 atom step/s1\text{\times}{10}^{3}\text{\,}\mathrm{atom}\text{\,}\mathrm{step}\text{/}\mathrm{s} using one typical CPU core, which is about two to three orders of magnitude slower than typical empirical potentials such as the Tersoff potential [38]. Therefore, even though MLPs are already faster than quantum-mechanical methods, it is still desirable to speed up MLPs as much as possible.

One approach to speed up MLPs is to use a huge number of CPUs and/or GPUs through message-passing information parallelisation. For example, using 27,360 V100 GPUs and the same number of CPU cores, the deep potential (DP) approach [17, 18, 19] has been used to simulate a 127-million-atom aluminum system with a speed of about 1.23×109 atom step/s1.23\text{\times}{10}^{9}\text{\,}\mathrm{atom}\text{\,}\mathrm{step}\text{/}\mathrm{s} [39]. However, such huge amounts of computational resources are not available to most researchers. More importantly, the performance of DP per V100 GPU is only about 4.5×104 atom step/s4.5\text{\times}{10}^{4}\text{\,}\mathrm{atom}\text{\,}\mathrm{step}\text{/}\mathrm{s}, which is about ten times slower than an empirical embedded atom method (EAM) potential [40] with a single typical CPU core.

A more economical approach is to optimize the formulation and implementation of the MLP itself so that it can attain a high computational speed using a reasonable amount of computational resources available to most researchers. To this end, we have developed a MLP called neuroevolution potential (NEP) [41, 42] within the gpumd package [43, 44] that can achieve a computational speed of about 1×107 atom step/s1\text{\times}{10}^{7}\text{\,}\mathrm{atom}\text{\,}\mathrm{step}\text{/}\mathrm{s} using a single V100 GPU, which is about ten times faster than the empirical Tersoff potential on a single typical CPU core.

In this paper, we present recent developments of the NEP approach that further improve its accuracy without reducing the efficiency. Specifically, we improve the radial functions in the atomic-environment descriptor by using a combination of basis functions, and add angular descriptor components with high-order correlations. The improved radial functions are better at distinguishing different atom types and lead to higher accuracy in multi-components systems. The added angular descriptor components with high-order correlations make the atomic-environment descriptor more complete and help to increase the regression accuracy.

Using a number of systems, including MgAlCu alloy, silicon with various phases, the azobenzene molecule, and carbon with various phases, we demonstrate the accuracy and efficiency of the latest NEP as implemented in the gpumd package. We compare with other state-of-the-art MLPs, including DP [17, 18, 19], Gaussian approximation potential (GAP) [8], momentum tensor potential (MTP) [30, 31], recursive embedded-atom neural network (REANN) [45, 32], and atomic cluster expansion (ACE) [46, 47, 48]. Through these comprehensive comparisons, we show that the NEP implementation in gpumd can achieve a computational speed that is far superior to other MLPs, under the condition of achieving an above-average accuracy. We present the algorithms for the efficient GPU implementation of NEP in great detail. Using a single GPU, such as an A100, one can use gpumd to simulate up to 10 million atoms on nanosecond time scales, which can only be achieved by using a huge amount of computational resources with other codes. The gpumd package makes large-scale, high-accuracy atomistic simulations available to a wide community instead of only a small number of institutions. In addition to these high-efficiency atomistic simulations, we also propose an effective active-learning scheme based on the latent space of the NEP model that can greatly reduce the computational burden of preparing training data.

Finally, we introduce the workflow for constructing and using NEPs through concrete examples in atomistic simulations of various materials properties, including lattice constant, elastic constants, stress-strain relation during tensile loading, structural properties during a melt-quench-anneal process, and thermal properties of amorphous structures. We also describe interfacing gpumd to Python via the gpymd and calorine packages.

II Theoretical formulations of the NEP approach

The first NEP, called NEP1, was proposed in Ref. 41. An improved version, called NEP2, was presented in Ref. 42. In the present paper, we further refine the NEP approach and introduce NEP3. In this section, we present NEP3 and discuss the differences to NEP1 and NEP2.

II.1 The neural-network model

Following Behler and Parrinello [49], the site energy of atom ii is taken as a function of the descriptor vector with NdesN_{\mathrm{des}} components, Ui​(𝐪)=Ui​({qνi}ν=1Ndes)U_{i}(\mathbf{q})=U_{i}\left(\{q^{i}_{\nu}\}_{\nu=1}^{N_{\mathrm{des}}}\right). We use a feedforward neural-network with a single hidden layer with NneuN_{\mathrm{neu}} neurons to model this function:

Ui=∑μ=1Nneuwμ(1)​tanh⁡(∑ν=1Ndeswμ​ν(0)​qνi−bμ(0))−b(1),U_{i}=\sum_{\mu=1}^{N_{\mathrm{neu}}}w^{(1)}_{\mu}\tanh\left(\sum_{\nu=1}^{N_{\mathrm{des}}}w^{(0)}_{\mu\nu}q^{i}_{\nu}-b^{(0)}_{\mu}\right)-b^{(1)}, (1)

where tanh⁡(x)\tanh(x) is the activation function in the hidden layer, 𝐰(0)\mathbf{w}^{(0)} is the connection weight matrix from the input layer (descriptor vector) to the hidden layer, 𝐰(1)\mathbf{w}^{(1)} is the connection weight vector from the hidden layer to the output node, which is the energy UiU_{i}, 𝐛(0)\mathbf{b}^{(0)} is the bias vector in the hidden layer, and b(1)b^{(1)} is the bias for node UiU_{i}. The total number of parameters in the neural network is thus (Ndes+2)​Nneu+1(N_{\mathrm{des}}+2)N_{\mathrm{neu}}+1. The descriptor vector is formed by juxtaposition of a number of components, including those with radial (distance) information only, which are called radial descriptor components, and those with both radial and angular information, which are called angular descriptor components. The descriptor is one of the most important aspects in MLPs [50, 51]. We discuss the radial and angular descriptor components below.

II.2 Radial descriptor components

There are nmaxR+1n_{\mathrm{max}}^{\mathrm{R}}+1 radial descriptor components and they are defined as

qni=∑j≠ign​(ri​j)with0≤n≤nmaxR,q^{i}_{n}=\sum_{j\neq i}g_{n}(r_{ij})\quad\text{with}\quad 0\leq n\leq n_{\mathrm{max}}^{\mathrm{R}}, (2)

where the summation runs over all the neighbors of atom ii within a certain cutoff distance. The functions gn​(ri​j)g_{n}(r_{ij}) depend on the distance ri​jr_{ij} only and are therefore called the radial functions. In NEP3, they are defined as a linear combination of NbasR+1N_{\mathrm{bas}}^{\mathrm{R}}+1 basis functions {fk​(ri​j)}k=0NbasR\{f_{k}(r_{ij})\}_{k=0}^{N_{\mathrm{bas}}^{\mathrm{R}}}:

gn​(ri​j)\displaystyle g_{n}(r_{ij}) =∑k=0NbasRcn​ki​jfk(ri​j)with\displaystyle=\sum_{k=0}^{N_{\mathrm{bas}}^{\mathrm{R}}}c^{ij}_{nk}f_{k}(r_{ij})\quad\text{with} (3)
fk​(ri​j)\displaystyle f_{k}(r_{ij}) =12​[Tk​(2​(ri​j/rcR−1)2−1)+1]​fc​(ri​j).\displaystyle=\frac{1}{2}\left[T_{k}\left(2\left(r_{ij}/r_{\mathrm{c}}^{\mathrm{R}}-1\right)^{2}-1\right)+1\right]f_{\mathrm{c}}(r_{ij}). (4)

Both nmaxRn_{\mathrm{max}}^{\mathrm{R}} and NbasRN_{\mathrm{bas}}^{\mathrm{R}} are tunable hyperparameters in NEP3 which, along with other ones, will be listed in subsubsection II.8.2. In subsection III.2, we will use a few examples to illustrate the judicious choice of the various hyperparameters in typical applications. Here, Tk​(x)T_{k}(x) is the kk-th order Chebyshev polynomial of the first kind and fc​(ri​j)f_{\mathrm{c}}(r_{ij}) is the cutoff function defined as

fc​(ri​j)={12​[1+cos⁡(π​ri​jrcR)]ri​j≤rcR0ri​j>rcRf_{\mathrm{c}}(r_{ij})=\begin{cases}\frac{1}{2}\left[1+\cos\left(\pi\frac{r_{ij}}{r_{\mathrm{c}}^{\mathrm{R}}}\right)\right]&r_{ij}\leq r_{\mathrm{c}}^{\mathrm{R}}\\ 0&r_{ij}>r_{\mathrm{c}}^{\mathrm{R}}\end{cases} (5)

Here, rcRr_{\mathrm{c}}^{\mathrm{R}} is the cutoff distance of the radial descriptor components. The expansion coefficients cn​ki​jc_{nk}^{ij} depend on nn and kk and also on the types of atoms ii and jj. Due to the summation over neighbors, the radial descriptor components defined above are invariant with respect to permutation of atoms of the same type.

If the material considered has NtypN_{\mathrm{typ}} atom types, the number of cn​ki​jc_{nk}^{ij} coefficients for the radial descriptor components is Ntyp2​(nmaxR+1)​(NbasR+1)N_{\mathrm{typ}}^{2}\left(n_{\mathrm{max}}^{\mathrm{R}}+1\right)\left(N_{\mathrm{bas}}^{\mathrm{R}}+1\right). In NEP2, each radial function is simply a basis function and cn​ki​jc_{nk}^{ij} is reduced to δn​k​cn​i​j\delta_{nk}c_{nij}, where δn​k\delta_{nk} is the Kronecker symbol and cn​i​jc_{nij} is defined in Ref. 42. In NEP2 and NEP3, these coefficients are trainable (similar to the parameters in the neural network), while in NEP1, we have used some fixed coefficients similar to previous works [14, 52]. In Ref. 42, we have shown that NEP2 is much more accurate than NEP1 for multi-component systems. In this paper, we will show that the accuracy of NEP3 for multi-component systems is further improved compared to NEP2.

II.3 Angular descriptor components

In NEP1 and NEP2 the angular descriptor components {qn​li}\{q^{i}_{nl}\} are taken as (1≤l≤lmax3​b1\leq l\leq l_{\mathrm{max}}^{\mathrm{3b}}):

qn​li=2​l+14​π​∑j≠i∑k≠ign​(ri​j)​gn​(ri​k)​Pl​(cos⁡θi​j​k),q^{i}_{nl}=\frac{2l+1}{4\pi}\sum_{j\neq i}\sum_{k\neq i}g_{n}(r_{ij})g_{n}(r_{ik})P_{l}(\cos\theta_{ijk}), (6)

where Pl​(cos⁡θi​j​k)P_{l}(\cos\theta_{ijk}) is the Legendre polynomial of order ll and θi​j​k\theta_{ijk} is the angle formed by the i​jij and i​kik bonds. The radial functions gn​(ri​j)g_{n}(r_{ij}) have the same forms as in Eq. (3), but can have a different cutoff distance rcAr_{\mathrm{c}}^{\mathrm{A}} and a different basis size NbasAN_{\mathrm{bas}}^{\mathrm{A}} than those in the radial descriptor components. Usually, it is beneficial to use rcA<rcRr_{\mathrm{c}}^{\mathrm{A}}<r_{\mathrm{c}}^{\mathrm{R}}, assuming that there is no directional dependence of the descriptor on some neighboring atoms that are sufficiently far away from the central atom. This reflects the physical intuition that interaction strength decreases both with distance and order. The radial descriptor components are relatively cheap to evaluate and one can thus use a relatively large radial cutoff distance rcRr_{\mathrm{c}}^{\mathrm{R}} (also relatively large nmaxRn_{\mathrm{max}}^{\mathrm{R}} and NbasRN_{\mathrm{bas}}^{\mathrm{R}}) combined with a relatively small angular cutoff distance rcAr_{\mathrm{c}}^{\mathrm{A}} (also relatively small nmaxAn_{\mathrm{max}}^{\mathrm{A}} and NbasAN_{\mathrm{bas}}^{\mathrm{A}}) to achieve a good balance between accuracy and speed. Note that message-passing could effectively increase the interaction range but is not necessarily an efficient way of describing long-range interactions [53]. Using a relatively large cutoff for the radial descriptor components is generally a more efficient way of incorporating long-ranged interactions, such as van-der-Waals (vdW) and screened Coulomb interactions, although it is incapable of describing genuine long-ranged interactions such as unscreened Coulomb interactions.

The expression (6) is not efficient for numerical evaluation due to the double summation over neighbors. An equivalent form that is more efficient for numerical evaluation can be obtained by using the addition theorem of the spherical harmonics as:

qn​li=∑m=−ll(−1)m​An​l​mi​An​l​(−m)i=∑m=0l(2−δ0​l)​|An​l​mi|2,q^{i}_{nl}=\sum_{m=-l}^{l}(-1)^{m}A^{i}_{nlm}A^{i}_{nl(-m)}=\sum_{m=0}^{l}(2-\delta_{0l})|A^{i}_{nlm}|^{2}, (7)

where

An​l​mi=∑j≠ign​(ri​j)​Yl​m​(θi​j,ϕi​j).A^{i}_{nlm}=\sum_{j\neq i}g_{n}(r_{ij})Y_{lm}(\theta_{ij},\phi_{ij}). (8)

Here, Yl​m​(θi​j,ϕi​j)Y_{lm}(\theta_{ij},\phi_{ij}) are the spherical harmonics as a function of the polar angle θi​j\theta_{ij} and the azimuthal angle ϕi​j\phi_{ij} for the position difference vector 𝒓i​j≡𝒓j−𝒓i\bm{r}_{ij}\equiv\bm{r}_{j}-\bm{r}_{i} from atom ii to atom jj. In Eq. (7), we have used the property An​l​(−m)=(−1)m​An​l​m∗A_{nl(-m)}=(-1)^{m}A_{nlm}^{\ast}, which follows from the property Yl⁡(−m)=(−1)m​Yl​m∗Y_{l(-m)}=(-1)^{m}Y_{lm}^{\ast}.

The angular descriptor components above are usually known as 3-body ones as in the ACE approach [46], although all the descriptor components are many-body in nature. For simplicity, we will use the ACE terminology. Higher-order angular descriptor components can be similarly constructed [46]. In NEP3, we add the following 4-body descriptor components (1≤l1≤l2≤l3≤lmax4​b1\leq l_{1}\leq l_{2}\leq l_{3}\leq l_{\mathrm{max}}^{\mathrm{4b}}):

qn​l1​l2​l3i=\displaystyle q^{i}_{nl_{1}l_{2}l_{3}}= ∑m1=−l1l1∑m2=−l2l2∑m3=−l3l3(l1l2l3m1m2m3)\displaystyle\sum_{m_{1}=-l_{1}}^{l_{1}}\sum_{m_{2}=-l_{2}}^{l_{2}}\sum_{m_{3}=-l_{3}}^{l_{3}}\left(\begin{array}[]{ccc}l_{1}&l_{2}&l_{3}\\ m_{1}&m_{2}&m_{3}\\ \end{array}\right)
×An​l1​m1i​An​l2​m2i​An​l3​m3i,\displaystyle\times A^{i}_{nl_{1}m_{1}}A^{i}_{nl_{2}m_{2}}A^{i}_{nl_{3}m_{3}}, (11)

and the following 5-body ones (1≤l1≤l2≤l3≤l4≤lmax5​b1\leq l_{1}\leq l_{2}\leq l_{3}\leq l_{4}\leq l_{\mathrm{max}}^{\mathrm{5b}}):

qin​l1​l2​l3​l4=∑m1=−l1l1∑m2=−l2l2∑m3=−l3l3∑m4=−l4l4\displaystyle q^{i}_{nl_{1}l_{2}l_{3}l_{4}}=\sum_{m_{1}=-l_{1}}^{l_{1}}\sum_{m_{2}=-l_{2}}^{l_{2}}\sum_{m_{3}=-l_{3}}^{l_{3}}\sum_{m_{4}=-l_{4}}^{l_{4}}
[l1l2l3l4m1m2m3m4]​An​l1​m1i​An​l2​m2i​An​l3​m3i​An​l4​m4i,\displaystyle\left[\begin{array}[]{cccc}l_{1}&l_{2}&l_{3}&l_{4}\\ m_{1}&m_{2}&m_{3}&m_{4}\\ \end{array}\right]A^{i}_{nl_{1}m_{1}}A^{i}_{nl_{2}m_{2}}A^{i}_{nl_{3}m_{3}}A^{i}_{nl_{4}m_{4}},

where (l1l2l3m1m2m3)\left(\begin{array}[]{ccc}l_{1}&l_{2}&l_{3}\\ m_{1}&m_{2}&m_{3}\\ \end{array}\right) are Wigner 3​j3j symbols and [46]

[l1l2l3l4m1m2m3m4]=∑L=max⁡{|l1−l2|,|l3−l4|}min⁡{|l1+l2|,|l3+l4|}∑M=−LL\displaystyle\left[\begin{array}[]{cccc}l_{1}&l_{2}&l_{3}&l_{4}\\ m_{1}&m_{2}&m_{3}&m_{4}\\ \end{array}\right]=\sum_{L=\max\{|l_{1}-l_{2}|,|l_{3}-l_{4}|\}}^{\min\{|l_{1}+l_{2}|,|l_{3}+l_{4}|\}}\sum_{M=-L}^{L}
(−1)M​(Ll1l2−Mm1m2)​(Ll3l4Mm3m4).\displaystyle(-1)^{M}\left(\begin{array}[]{ccc}L&l_{1}&l_{2}\\ -M&m_{1}&m_{2}\\ \end{array}\right)\left(\begin{array}[]{ccc}L&l_{3}&l_{4}\\ M&m_{3}&m_{4}\\ \end{array}\right).

We can consider higher-order terms [46], but to keep a balance between accuracy and speed, we only consider those up to fifth order. Recent work suggests that 4-body interactions are difficult to learn using only 3-body correlations [54] and that using only up to 4-body correlations can still yield identical results for different configurations of simple molecules [55]. One can formally achieve completeness in the MTP formalism [30], the ACE formalism [46, 48], and other related ones [56, 57, 58], but in practice, there is always a truncation of the descriptor size and a balance must be struck between accuracy and speed. This balance is one of the most important guidelines for the development of NEP in gpumd.

II.4 Explicit expressions of the angular descriptor components

The angular descriptor components are quite complicated and care must be taken to achieve an efficient implementation. There have been some implementations of the ACE approach combined with linear regression, where it has been found that recursive evaluation can lead to much higher efficiency [48, 35, 36]. In our GPU implementation of NEP3 with the ACE-like descriptor components, we find it crucial to derive the relevant expressions as explicitly as possible as it allows us to reduce the number of terms to be evaluated thanks to symmetry considerations.

To facilitate the following presentation, we define a series of summations that are used to express the angular descriptor components (0≤n≤nmaxA0\leq n\leq n_{\mathrm{max}}^{\mathrm{A}} and 0≤k≤230\leq k\leq 23):

Sn,k=∑j≠ign​(ri​j)ri​jn​bk​(xi​j,yi​j,zi​j).S_{n,k}=\sum_{j\neq i}\frac{g_{n}(r_{ij})}{r_{ij}^{n}}b_{k}(x_{ij},y_{ij},z_{ij}). (20)

The functions bk​(xi​j,yi​j,zi​j)b_{k}(x_{ij},y_{ij},z_{ij}) here are zi​jz_{ij}, xi​jx_{ij}, yi​jy_{ij}, 3​zi​j2−ri​j23z_{ij}^{2}-r_{ij}^{2}, xi​j​zi​jx_{ij}z_{ij}, yi​j​zi​jy_{ij}z_{ij}, xi​j2−yi​j2x_{ij}^{2}-y_{ij}^{2}, 2​xi​j​yi​j2x_{ij}y_{ij}, (5​zi​j2−3​ri​j2)​zi​j(5z_{ij}^{2}-3r_{ij}^{2})z_{ij}, (5​zi​j2−ri​j2)​xi​j(5z_{ij}^{2}-r_{ij}^{2})x_{ij}, (5​zi​j2−ri​j2)​yi​j(5z_{ij}^{2}-r_{ij}^{2})y_{ij}, (xi​j2−yi​j2)​zi​j(x_{ij}^{2}-y_{ij}^{2})z_{ij}, 2​xi​j​yi​j​zi​j2x_{ij}y_{ij}z_{ij}, (xi​j2−3​yi​j2)​xi​j(x_{ij}^{2}-3y_{ij}^{2})x_{ij}, (3​xi​j2−yi​j2)​yi​j(3x_{ij}^{2}-y_{ij}^{2})y_{ij}, (35​zi​j2−30​ri​j2)​zi​j2+3​ri​j4(35z_{ij}^{2}-30r_{ij}^{2})z_{ij}^{2}+3r_{ij}^{4}, (7​zi​j2−3​ri​j2)​xi​j​zi​j(7z_{ij}^{2}-3r_{ij}^{2})x_{ij}z_{ij}, (7​zi​j2−3​ri​j2)​yi​j​zi​j(7z_{ij}^{2}-3r_{ij}^{2})y_{ij}z_{ij}, (7​zi​j2−ri​j2)​(xi​j2−yi​j2)(7z_{ij}^{2}-r_{ij}^{2})(x_{ij}^{2}-y_{ij}^{2}), (7​zi​j2−ri​j2)​2​xi​j​yi​j(7z_{ij}^{2}-r_{ij}^{2})2x_{ij}y_{ij}, (xi​j2−3​yi​j2)​xi​j​zi​j(x_{ij}^{2}-3y_{ij}^{2})x_{ij}z_{ij}, (3​xi​j2−yi​j2)​yi​j​zi​j(3x_{ij}^{2}-y_{ij}^{2})y_{ij}z_{ij}, (xi​j2−yi​j2)2−4​xi​j2​yi​j2(x_{ij}^{2}-y_{ij}^{2})^{2}-4x_{ij}^{2}y_{ij}^{2}, and 4​(xi​j2−yi​j2)​xi​j​yi​j4(x_{ij}^{2}-y_{ij}^{2})x_{ij}y_{ij} from k=0k=0 to k=23k=23. With these summations, we can write the 3-body angular descriptor components up to lmax3​b=4l_{\mathrm{max}}^{\mathrm{3b}}=4 explicitly as:

qn​1i=14​3π​Sn,02+2⁤14​32​π​(Sn,12+Sn,22)≡∑k=02Ck3​b​Sn,k2;\displaystyle q^{i}_{n1}=\frac{1}{4}\frac{3}{\pi}S_{n,0}^{2}+2\frac{1}{4}\frac{3}{2\pi}(S_{n,1}^{2}+S_{n,2}^{2})\equiv\sum_{k=0}^{2}C^{\mathrm{3b}}_{k}S_{n,k}^{2}; (21)
qn​2i\displaystyle q^{i}_{n2} =116​5π​Sn,32+2⁤14​152​π​(Sn,42+Sn,52)+2⁤116​152​π​(Sn,62+Sn,72)\displaystyle=\frac{1}{16}\frac{5}{\pi}S_{n,3}^{2}+2\frac{1}{4}\frac{15}{2\pi}(S_{n,4}^{2}+S_{n,5}^{2})+2\frac{1}{16}\frac{15}{2\pi}(S_{n,6}^{2}+S_{n,7}^{2})
≡∑k=37Ck3​b​Sn,k2;\displaystyle\equiv\sum_{k=3}^{7}C^{\mathrm{3b}}_{k}S_{n,k}^{2}; (22)
qn​3i\displaystyle q^{i}_{n3} =116​7π​Sn,82+2⁤164​21π​(Sn,92+Sn,102)\displaystyle=\frac{1}{16}\frac{7}{\pi}S_{n,8}^{2}+2\frac{1}{64}\frac{21}{\pi}(S_{n,9}^{2}+S_{n,10}^{2})
+2⁤116​1052​π​(Sn,112+Sn,122)+2⁤164​35π​(Sn,132+Sn,142)\displaystyle+2\frac{1}{16}\frac{105}{2\pi}(S_{n,11}^{2}+S_{n,12}^{2})+2\frac{1}{64}\frac{35}{\pi}(S_{n,13}^{2}+S_{n,14}^{2})
≡∑k=814Ck3​b​Sn,k2;\displaystyle\equiv\sum_{k=8}^{14}C^{\mathrm{3b}}_{k}S_{n,k}^{2}; (23)
qn​4i\displaystyle q^{i}_{n4} =9256​1π​Sn,152+2⁤964​5π​(Sn,162+Sn,172)\displaystyle=\frac{9}{256}\frac{1}{\pi}S_{n,15}^{2}+2\frac{9}{64}\frac{5}{\pi}(S_{n,16}^{2}+S_{n,17}^{2})
+2⁤964​52​π​(Sn,182+Sn,192)+2⁤964​35π​(Sn,202+Sn,212)\displaystyle+2\frac{9}{64}\frac{5}{2\pi}(S_{n,18}^{2}+S_{n,19}^{2})+2\frac{9}{64}\frac{35}{\pi}(S_{n,20}^{2}+S_{n,21}^{2})
+2⁤9256​352​π​(Sn,222+Sn,232)≡∑k=1523Ck3​b​Sn,k2,\displaystyle+2\frac{9}{256}\frac{35}{2\pi}(S_{n,22}^{2}+S_{n,23}^{2})\equiv\sum_{k=15}^{23}C^{\mathrm{3b}}_{k}S_{n,k}^{2}, (24)

whereby we defined the 3-body coefficients {Ck3​b}k=023\{C^{\mathrm{3b}}_{k}\}_{k=0}^{23}.

For 4-body angular descriptor components, we only consider the case of l1=l2=l3l_{1}=l_{2}=l_{3} and up to lmax4​b=2l^{\mathrm{4b}}_{\mathrm{max}}=2. It turns out that qn​111i=qn​333i=0q^{i}_{n111}=q^{i}_{n333}=0. Therefore, there is no difference between lmax4​b=2l^{\mathrm{4b}}_{\mathrm{max}}=2 and lmax4​b=3l^{\mathrm{4b}}_{\mathrm{max}}=3. Then we only have the case of l1=l2=l3=2l_{1}=l_{2}=l_{3}=2,

qn​222i\displaystyle q^{i}_{n222} =−235​An​20i​An​20i​An​20i\displaystyle=-\sqrt{\frac{2}{35}}A^{i}_{n20}A^{i}_{n20}A^{i}_{n20}
+6​170​An​20i​An​21i​An​2​(−1)i\displaystyle+6\sqrt{\frac{1}{70}}A^{i}_{n20}A^{i}_{n21}A^{i}_{n2(-1)}
+6​235​An​20i​An​22i​An​2​(−2)i\displaystyle+6\sqrt{\frac{2}{35}}A^{i}_{n20}A^{i}_{n22}A^{i}_{n2(-2)}
−6​335​An​21i​An​21i​An​2​(−2)i.\displaystyle-6\sqrt{\frac{3}{35}}A^{i}_{n21}A^{i}_{n21}A^{i}_{n2(-2)}. (25)

The root-rational-fraction package [59] has been used to obtain analytical expressions of the various Wigner 3​j3j symbols. With some algebra, we have

qn​222i\displaystyle q^{i}_{n222} =−235​(14​5π)3​Sn,33\displaystyle=-\sqrt{\frac{2}{35}}\left(\frac{1}{4}\sqrt{\frac{5}{\pi}}\right)^{3}S_{n,3}^{3}
−6​170​(14​5π)​(12​152​π)2​Sn,3​(Sn,42+Sn,52)\displaystyle-6\sqrt{\frac{1}{70}}\left(\frac{1}{4}\sqrt{\frac{5}{\pi}}\right)\left(\frac{1}{2}\sqrt{\frac{15}{2\pi}}\right)^{2}S_{n,3}(S_{n,4}^{2}+S_{n,5}^{2})
+6​235​(14​5π)​(14​152​π)2​Sn,3​(Sn,62+Sn,72)\displaystyle+6\sqrt{\frac{2}{35}}\left(\frac{1}{4}\sqrt{\frac{5}{\pi}}\right)\left(\frac{1}{4}\sqrt{\frac{15}{2\pi}}\right)^{2}S_{n,3}(S_{n,6}^{2}+S_{n,7}^{2})
+6​335​(12​152​π)2​(14​152​π)​Sn,6​(Sn,52−Sn,42)\displaystyle+6\sqrt{\frac{3}{35}}\left(\frac{1}{2}\sqrt{\frac{15}{2\pi}}\right)^{2}\left(\frac{1}{4}\sqrt{\frac{15}{2\pi}}\right)S_{n,6}(S_{n,5}^{2}-S_{n,4}^{2})
−12​335​(12​152​π)2​(14​152​π)​Sn,4​Sn,5​Sn,7\displaystyle-12\sqrt{\frac{3}{35}}\left(\frac{1}{2}\sqrt{\frac{15}{2\pi}}\right)^{2}\left(\frac{1}{4}\sqrt{\frac{15}{2\pi}}\right)S_{n,4}S_{n,5}S_{n,7}
≡C04​b​Sn,33+C14​b​Sn,3​(Sn,42+Sn,52)\displaystyle\equiv C^{\mathrm{4b}}_{0}S_{n,3}^{3}+C^{\mathrm{4b}}_{1}S_{n,3}(S_{n,4}^{2}+S_{n,5}^{2})
+C24​b​Sn,3​(Sn,62+Sn,72)\displaystyle+C^{\mathrm{4b}}_{2}S_{n,3}(S_{n,6}^{2}+S_{n,7}^{2})
+C34​b​Sn,6​(Sn,52−Sn,42)+C44​b​Sn,4​Sn,5​Sn,7,\displaystyle+C^{\mathrm{4b}}_{3}S_{n,6}(S_{n,5}^{2}-S_{n,4}^{2})+C^{\mathrm{4b}}_{4}S_{n,4}S_{n,5}S_{n,7}, (26)

whereby we defined the 4-body coefficients {Ck4​b}k=04\{C^{\mathrm{4b}}_{k}\}_{k=0}^{4}.

For 5-body descriptors, we only consider up to l1=l2=l3=l4=1l_{1}=l_{2}=l_{3}=l_{4}=1,

qn​1111i\displaystyle q^{i}_{n1111} =715​(An​10i)4−2815​(An​10i)2​An​11i​An​1​(−1)i\displaystyle=\frac{7}{15}\left(A^{i}_{n10}\right)^{4}-\frac{28}{15}\left(A^{i}_{n10}\right)^{2}A^{i}_{n11}A^{i}_{n1(-1)}
+2815​(An​11i)2​(An​1​(−1)i)2\displaystyle+\frac{28}{15}\left(A^{i}_{n11}\right)^{2}\left(A^{i}_{n1(-1)}\right)^{2}
=720​π​Sn,04+710​π​Sn,02​(Sn,12+Sn,22)\displaystyle=\frac{7}{20\pi}S_{n,0}^{4}+\frac{7}{10\pi}S_{n,0}^{2}(S_{n,1}^{2}+S_{n,2}^{2})
+710​π​(Sn,12+Sn,22)2\displaystyle+\frac{7}{10\pi}(S_{n,1}^{2}+S_{n,2}^{2})^{2}
≡C05​b​Sn,04+C15​b​Sn,02​(Sn,12+Sn,22)\displaystyle\equiv C^{\mathrm{5b}}_{0}S_{n,0}^{4}+C^{\mathrm{5b}}_{1}S_{n,0}^{2}(S_{n,1}^{2}+S_{n,2}^{2})
+C25​b​(Sn,12+Sn,22)2,\displaystyle+C^{\mathrm{5b}}_{2}(S_{n,1}^{2}+S_{n,2}^{2})^{2}, (27)

whereby we defined the 5-body coefficients {Ck5​b}k=02\{C^{\mathrm{5b}}_{k}\}_{k=0}^{2}.

In our implementation, the 3-body coefficients {Ck3​b}k=023\{C^{\mathrm{3b}}_{k}\}_{k=0}^{23}, 4-body coefficients {Ck4​b}k=04\{C^{\mathrm{4b}}_{k}\}_{k=0}^{4}, and 5-body coefficients {Ck5​b}k=02\{C^{\mathrm{5b}}_{k}\}_{k=0}^{2} are pre-computed. This is crucial for obtaining high computational performance.

We can now enumerate the descriptor vector length. There are (nmaxR+1)(n_{\mathrm{max}}^{\mathrm{R}}+1) radial descriptor components, (nmaxA+1)​lmax3​b(n_{\mathrm{max}}^{\mathrm{A}}+1)l_{\mathrm{max}}^{\mathrm{3b}} 3-body descriptor components, (nmaxA+1)(n_{\mathrm{max}}^{\mathrm{A}}+1) 4-body descriptor components, and (nmaxA+1)(n_{\mathrm{max}}^{\mathrm{A}}+1) 5-body descriptor components. Therefore, we have

Ndes=(nmaxR+1)+(nmaxA+1)​(lmax3​b+2)N_{\mathrm{des}}=\left(n_{\mathrm{max}}^{\mathrm{R}}+1\right)+\left(n_{\mathrm{max}}^{\mathrm{A}}+1\right)\left(l_{\mathrm{max}}^{\mathrm{3b}}+2\right) (28)

in NEP3 if we include both the 4-body and the 5-body descriptor components.

II.5 Force, virial, and heat current expressions

As stressed in Ref. 44, we need to derive an explicit expression of the partial force [60] for an efficient GPU implementation.

The partial force is

∂Ui∂𝒓i​j=\displaystyle\frac{\partial U_{i}}{\partial\bm{r}_{ij}}= ∑n=0nmaxR∂Ui∂qni​∂qni∂𝒓i​j+∑n=0nmaxA∑l=1lmax3​b∂Ui∂qn​li​∂qn​li∂𝒓i​j\displaystyle\sum_{n=0}^{n_{\mathrm{max}}^{\mathrm{R}}}\frac{\partial U_{i}}{\partial q^{i}_{n}}\frac{\partial q^{i}_{n}}{\partial\bm{r}_{ij}}+\sum_{n=0}^{n_{\mathrm{max}}^{\mathrm{A}}}\sum_{l=1}^{l^{\mathrm{3b}}_{\mathrm{max}}}\frac{\partial U_{i}}{\partial q^{i}_{nl}}\frac{\partial q^{i}_{nl}}{\partial\bm{r}_{ij}}
+\displaystyle+ ∑n=0nmaxA∑l=1lmax4​b∂Ui∂qn​l​l​li​∂qn​l​l​li∂𝒓i​j+∑n=0nmaxA∑l=1lmax5​b∂Ui∂qn​l​l​l​li​∂qn​l​l​l​li∂𝒓i​j.\displaystyle\sum_{n=0}^{n_{\mathrm{max}}^{\mathrm{A}}}\sum_{l=1}^{l^{\mathrm{4b}}_{\mathrm{max}}}\frac{\partial U_{i}}{\partial q^{i}_{nlll}}\frac{\partial q^{i}_{nlll}}{\partial\bm{r}_{ij}}+\sum_{n=0}^{n_{\mathrm{max}}^{\mathrm{A}}}\sum_{l=1}^{l^{\mathrm{5b}}_{\mathrm{max}}}\frac{\partial U_{i}}{\partial q^{i}_{nllll}}\frac{\partial q^{i}_{nllll}}{\partial\bm{r}_{ij}}. (29)

Because all the relevant functions here are analytical, it is straightforward to derive explicit expressions for all the partial derivatives in the equation above.

With the partial force available, the total force acting on atom ii from atom jj can be computed as [60]

𝑭i​j=∂Ui∂𝒓i​j−∂Uj∂𝒓j​i,\bm{F}_{ij}=\frac{\partial U_{i}}{\partial\bm{r}_{ij}}-\frac{\partial U_{j}}{\partial\bm{r}_{ji}}, (30)

which respects Newton’s third law 𝑭i​j=−𝑭j​i\bm{F}_{ij}=-\bm{F}_{ji}. The total force acting on atom ii from all the neighboring atoms is thus

𝑭i=∑j≠i𝑭i​j.\bm{F}_{i}=\sum_{j\neq i}\bm{F}_{ij}. (31)

From the partial force, one can define the per-atom virial [60, 61]

𝐖i=∑j≠i𝒓i​j⊗∂Uj∂𝒓j​i.\mathbf{W}_{i}=\sum_{j\neq i}\bm{r}_{ij}\otimes\frac{\partial U_{j}}{\partial\bm{r}_{ji}}. (32)

By contracting the per-atom virial above with the velocity 𝒗i\bm{v}_{i}, one can then obtain the per-atom heat current [60, 61]:

𝑱i\displaystyle\bm{J}_{i} =𝐖i⋅𝒗i=∑j≠i(𝒓i​j⊗∂Uj∂𝒓j​i)⋅𝒗i\displaystyle=\mathbf{W}_{i}\cdot\bm{v}_{i}=\sum_{j\neq i}\left(\bm{r}_{ij}\otimes\frac{\partial U_{j}}{\partial\bm{r}_{ji}}\right)\cdot\bm{v}_{i}
=∑j≠i𝒓i​j​(∂Uj∂𝒓j​i⋅𝒗i).\displaystyle=\sum_{j\neq i}\bm{r}_{ij}\left(\frac{\partial U_{j}}{\partial\bm{r}_{ji}}\cdot\bm{v}_{i}\right). (33)

The total heat current in the system is the sum of the per-atom contributions:

𝑱=∑i𝑱i=∑i∑j≠i𝒓i​j​(∂Uj∂𝒓j​i⋅𝒗i).\bm{J}=\sum_{i}\bm{J}_{i}=\sum_{i}\sum_{j\neq i}\bm{r}_{ij}\left(\frac{\partial U_{j}}{\partial\bm{r}_{ji}}\cdot\bm{v}_{i}\right). (34)

By an exchange of dummy indices, we can also write Eq. (34) as

𝑱=−∑i∑j≠i𝒓i​j(∂Ui∂𝒓i​j⋅𝒗j).\bm{J}=-\sum_{i}\sum_{j\neq i}\bm{r}_{ij}\left(\frac{\partial U_{i}}{\partial\bm{r}_{ij}}\cdot\bm{v}_{j}\right). (35)

Both Eqs. (34) and (35) can be used in the Green-Kubo method for thermal conductivity calculations, but Eq. (34) is a more convenient form for the homogeneous non-equilibrium molecular dynamics (HNEMD) method and the related spectral decomposition method, as it does not involved the velocities 𝒗j\bm{v}_{j} of the neighboring atoms jj for a given atom ii.

The heat current expressions above apply to all the interatomic potentials implemented in gpumd. In all the cases, the validity of the heat current expressions has been numerically confirmed in terms of energy conservation [62, 63, 64, 65]. Equation (35) has been recently used in an on-the-fly MLP [66]. Since the only assumptions for the derivations [60] of these expressions are the locality properties of the interatomic potentials, our heat current expressions generally apply to the multi-body cluster potentials as considered in Ref. 67, as we show in Appendix A.

II.6 Loss function and training algorithm

We use the separable natural evolution strategy (SNES) [68] to optimize the free parameters in NEP. We denote a set of parameters as a vector 𝐳\mathbf{z}, whose dimension is the total number of parameters NparN_{\mathrm{par}}. In NEP1,

Npar=(Ndes+2)​Nneu+1,N_{\mathrm{par}}=(N_{\mathrm{des}}+2)N_{\mathrm{neu}}+1, (36)

which is the same for both single-component and multi-component systems. In NEP2, we added trainable parameters to the descriptor for multi-component systems and we have

Npar=(Ndes+2)​Nneu+1+Ntyp2​(nmaxR+nmaxA+2),N_{\mathrm{par}}=(N_{\mathrm{des}}+2)N_{\mathrm{neu}}+1+N_{\mathrm{typ}}^{2}(n^{\mathrm{R}}_{\mathrm{max}}+n^{\mathrm{A}}_{\mathrm{max}}+2), (37)

if Ntyp>1N_{\mathrm{typ}}>1, and the same NparN_{\mathrm{par}} as in NEP1 if Ntyp=1N_{\mathrm{typ}}=1. In NEP3, the number of trainable parameters in the descriptor is increased for both single and multi-component systems and we have

Npar\displaystyle N_{\mathrm{par}} =(Ndes+2)​Nneu+1\displaystyle=(N_{\mathrm{des}}+2)N_{\mathrm{neu}}+1
+Ntyp2​(nmaxR+1)​(NbasR+1)\displaystyle+N_{\mathrm{typ}}^{2}(n^{\mathrm{R}}_{\mathrm{max}}+1)(N_{\mathrm{bas}}^{\mathrm{R}}+1)
+Ntyp2​(nmaxA+1)​(NbasA+1).\displaystyle+N_{\mathrm{typ}}^{2}(n^{\mathrm{A}}_{\mathrm{max}}+1)(N_{\mathrm{bas}}^{\mathrm{A}}+1). (38)

We can formally express the loss function as a function of the free parameters,

L=L⁡(𝐳),L=L(\mathbf{z}), (39)

and express the training process as a real-valued optimization problem:

𝐳∗=arg⁡min𝐳⁡L⁡(𝐳),\mathbf{z}^{\ast}=\arg\min_{\mathbf{z}}L(\mathbf{z}), (40)

where 𝐳∗\mathbf{z}^{\ast} is an optimal set of parameters.

The total loss function is defined as a weighted sum of all these individual loss functions:

L⁡(𝐳)\displaystyle L(\mathbf{z}) =λe​(1Nstr​∑n=1Nstr(UNEP​(n,𝐳)−Utar​(n))2)1/2\displaystyle=\lambda_{\mathrm{e}}\left(\frac{1}{N_{\mathrm{str}}}\sum_{n=1}^{N_{\mathrm{str}}}\left(U^{\mathrm{NEP}}(n,\mathbf{z})-U^{\mathrm{tar}}(n)\right)^{2}\right)^{1/2}
+λf​(13​N​∑i=1N(𝑭iNEP​(𝐳)−𝑭itar)2)1/2\displaystyle+\lambda_{\mathrm{f}}\left(\frac{1}{3N}\sum_{i=1}^{N}\left(\bm{F}_{i}^{\mathrm{NEP}}(\mathbf{z})-\bm{F}_{i}^{\mathrm{tar}}\right)^{2}\right)^{1/2}
+λv​(16​Nstr​∑n=1Nstr∑μ​ν(Wμ​νNEP​(n,𝐳)−Wμ​νtar​(n))2)1/2\displaystyle+\lambda_{\mathrm{v}}\left(\frac{1}{6N_{\mathrm{str}}}\sum_{n=1}^{N_{\mathrm{str}}}\sum_{\mu\nu}\left(W_{\mu\nu}^{\mathrm{NEP}}(n,\mathbf{z})-W_{\mu\nu}^{\mathrm{tar}}(n)\right)^{2}\right)^{1/2}
+λ11Npar∑n=1Npar|zn|\displaystyle+\lambda_{1}\frac{1}{N_{\mathrm{par}}}\sum_{n=1}^{N_{\mathrm{par}}}|z_{n}|
+λ2​(1Npar​∑n=1Nparzn2)1/2,\displaystyle+\lambda_{2}\left(\frac{1}{N_{\mathrm{par}}}\sum_{n=1}^{N_{\mathrm{par}}}z_{n}^{2}\right)^{1/2}, (41)

where NstrN_{\mathrm{str}} is the number of structures in the training data set (if using a full batch) or the number of structures in a mini-batch and NN is the total number of atoms in these structures. UNEP​(n,𝐳)U^{\mathrm{NEP}}(n,\mathbf{z}) and Wμ​νNEP​(n,𝐳)W_{\mu\nu}^{\mathrm{NEP}}(n,\mathbf{z}) are per-atom energy and virial tensor predicted by the NEP with parameters 𝐳\mathbf{z} for the nn-th structure, and 𝑭iNEP​(𝐳)\bm{F}_{i}^{\mathrm{NEP}}(\mathbf{z}) is the predicted force for the iith atom. Utar​(n)U^{\mathrm{tar}}(n), Wμ​νtar​(n)W_{\mu\nu}^{\mathrm{tar}}(n) and 𝑭itar\bm{F}_{i}^{\mathrm{tar}} are the corresponding target values. That is, the loss functions for energy, force, and virial are defined as their root mean square errors (RMSEs) between the current NEP predictions and the target values. The last two terms represent ℓ1\ell_{1} and ℓ2\ell_{2} regularization. The weights λe\lambda_{\mathrm{e}}, λf\lambda_{\mathrm{f}}, λv\lambda_{\mathrm{v}}, λ1\lambda_{1}, and λ2\lambda_{2} are tunable hyper-parameters. When calculating the loss function, we use the following units: eV/atom\mathrm{eV}\text{/}\mathrm{atom} for energy and virial and eV/Å\mathrm{eV}\text{/}\mathrm{\text{Å}} for force components.

The SNES [68] we use for optimizing Eq. (41) is a principled approach to real-valued evolutionary optimization by following the natural gradient of the loss function to update a search distribution (a mean value and a variance for each trainable parameter) for a population of solutions. It is a derivative-free black-box optimizer and thus does not require the loss function to have any analytical properties. An explicit workflow of the training algorithm has been presented in Ref. 41.

Refer to caption
Figure 1: The gpumd package includes two executables, gpumd and nep, which are represented by the two black boxes. The nep program can be used for training NEP models and the gpumd program can be used to perform atomistic simulations using the trained potentials. See text for details.

II.7 GPU implementation

The NEP approach is implemented in the open-source gpumd package, which is a general-purpose molecular dynamics (MD) simulation package fully implemented on GPUs. It currently supports only Nvidia GPUs and the programming language is CUDA C++. In this section as well as Appendix B, we present the detailed algorithms for our CUDA implementation of the NEP approach.

Similar to many other MLPs, NEP is a many-body potential and is very similar to empirical many-body potentials such as the EAM potential [40] and the Tersoff potential [38]. Specifically, the radial descriptor part resembles an EAM potential and the angular descriptor part resembles a Tersoff potential. Therefore, our CUDA implementation of the NEP approach follows the established efficient scheme for empirical many-body potentials [44].

The overall strategy of our CUDA implementation of the NEP approach is to use a few CUDA kernels only, which ensures a high degree of parallelism and high arithmetic intensity, both of which are crucial for achieving high performance in CUDA programming. In all the CUDA kernels, one CUDA thread is assigned to one atom, i.e., there is a one-to-one correspondence between atoms and CUDA threads. The descriptor vector and the various per-atom quantities, including the site energy, force, and virial for one atom, are all calculated within one CUDA thread. This also includes the application of the neural network as represented by Eq. (1). Appendix B.1 shows the CUDA __device__ function (to be called in the first CUDA kernel function as discussed below) evaluating Eq.  (1) as well as {∂Ui/∂qνi}\{\partial U_{i}/\partial q^{i}_{\nu}\}. One can see that the feedforward neural network used here is an analytical multi-variable scalar function.

The evaluation of NEP related quantities (energy, force, and virial) requires 4 CUDA kernels only:

  1. 1.

    The first CUDA kernel calculates the whole descriptor vector and applies the neural network to obtain the site energy and the derivatives of the energy with respect to the descriptor components, see Algorithm 1 in Appendix B.

  2. 2.

    The second CUDA kernel calculates the force and virial related to the radial descriptor components, see Algorithm 2 in Appendix B.

  3. 3.

    The third CUDA kernel calculates the partial force related to the angular descriptor components, see Algorithm 3 in Appendix B.

  4. 4.

    The fourth CUDA kernel calculates the force and virial related to the angular descriptor components, see Algorithm 4 in Appendix B.

In these CUDA kernels, the inputs and outputs related to the atoms (position, energy, force, and virial) are all in double precision, but the internal calculations within the CUDA kernels are mostly in single precision. This is an effective mixed-precision approach which can keep a good balance between accuracy and efficiency that has been adopted in many other GPU-accelerated atomistic simulation packages [69, 70, 71].

II.8 Training and using NEPs

II.8.1 Overview of the GPUMD package

The gpumd package [43, 44] can be used to train NEPs and use them in atomistic simulations. gpumd can be compiled and used in both Linux and Windows, provided that a CUDA development environment and a CUDA-enabled GPU are available. After compilation, one obtains the nep and gpumd executable, which can be used to train and use NEPs, respectively. Figure 1 provides a schematic overview of the workflow of training and using NEPs.

The gpumd package started from a minimal CUDA code implementing only simple pairwise potentials and thermal conductivity calculations [43]. Gradually, empirical many-body potentials were implemented using the unique formalism we proposed [60, 44], including the Tersoff [38] and Tersoff-like [72] potentials, the Stillinger-Weber potential [73], and EAM potentials [40]. Recently, support was also added for machine-learned force constant potentials constructed using the hiphive package [74] [65]. The most recent addition are the various versions of NEP as developed in the previous papers [41, 42] as well as the current one.

Apart from supporting the above important interatomic potentials, gpumd also supports many statistical ensembles, including the NVE (microcanonical), NVT (isothermal), and NPT (isothermal-isobaric) ensembles. For the NVT ensemble, it has options for the Berendsen thermostat [75], the Nosé-Hoover chain thermostat [76, 77, 78], the Bussi-Donadio-Parrinello thermostat [79], and the Langevin thermostat in different flavors [80, 81]. For the NPT ensemble, gpumd supports the classical Berendsen barostat [75] and the recently proposed stochastic cell-rescaling barostat by Bernetti and Bussi [82].

gpumd has been mainly used to study thermal transport. It supports the equilibrium MD method based on the Green-Kubo relation [83, 84], the non-equilibrium MD method using two or more thermostats (preferably the Langevin thermostats [85]), and the HNEMD method using a homogeneous driving force [86, 87]. The thermal transport coefficients, either thermal conductance or thermal conductivity, can be decomposed both spatially and spectrally [63, 87, 61]. gpumd has been used to establish the best practices [85, 88] in MD simulations of thermal transport. It has also been used to study the particular thermal transport properties of specific materials, including various two-dimensional (2D) materials [89, 90, 91, 92, 93, 64, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103], vdW structures based on 2D materials [104, 105, 106, 107, 108, 109, 110], and quasi-one-dimensional materials [111, 112, 113]. There are applications focused on revealing unique phonon transport mechanisms [114, 115, 116, 117, 118, 119, 120]. The high efficiency of gpumd also enabled high-throughput thermal transport simulations that were used as training/testing data for machine learning models of interfacial thermal transport [121].

Although previous studies using gpumd have been mostly focused on thermal transport properties, gpumd has already been developed into a general-purpose atomistic simulation package. In Sect. V we will showcase a series of typical atomistic simulations using a NEP model for carbon systems trained in this paper.

II.8.2 Training a NEP model

To train a NEP model, one needs to prepare three input files: train.in, test.in, and nep.in. The first two contain the training and the testing data, respectively. The files train.in and test.in have the same data format, the only difference being that the data in train.in will be used for training and those in test.in will be used for testing. The file train.in (test.in) contains the following data: the number of structures in the training (testing) set, the reference energy, reference virial tensor (optional), and cell metric for each structure, as well as the chemical symbol, position vector, and reference force vector for each atom in each structure. For the specific data format, we refer the reader to the gpumd manual.

The file nep.in contains hyper-parameters that define the NEP model and control the training process. In this file, one can choose the NEP version (currently NEP2 or NEP3), specify the number of atom types and their chemical symbols, the cutoff distances rcRr^{\mathrm{R}}_{\mathrm{c}} and rcAr^{\mathrm{A}}_{\mathrm{c}}, the radial function parameters (nmaxRn^{\mathrm{R}}_{\mathrm{max}}, nmaxAn^{\mathrm{A}}_{\mathrm{max}}, NbasRN^{\mathrm{R}}_{\mathrm{bas}}, and NbasAN^{\mathrm{A}}_{\mathrm{bas}}), the angular expansion parameters (lmax3​bl^{\mathrm{3b}}_{\mathrm{max}}, lmax4​bl^{\mathrm{4b}}_{\mathrm{max}}, and lmax5​bl^{\mathrm{5b}}_{\mathrm{max}}), the weights for the different terms in the loss function (λe\lambda_{\mathrm{e}}, λf\lambda_{\mathrm{f}}, λv\lambda_{\mathrm{v}}, λ1\lambda_{1}, and λ2\lambda_{2}), the number of neurons NneuN_{\mathrm{neu}} in the hidden layer of the neural network, the batch size NbatN_{\mathrm{bat}} (number of structures in one batch), the population size NpopN_{\mathrm{pop}}, and the number of generations NgenN_{\mathrm{gen}} in the SNES training algorithm. Details concerning the nep.in file are presented in the gpumd manual.

During the training process, predicted energy, force, virial values, the various terms of the loss function (both for the training and the testing data sets), the potential file, and a file used for restarting are continuously updated. Further details concerning the output files of the nep executable can be found in the gpumd manual.

II.8.3 Using a NEP

The potential file nep.txt contains all the information that constitutes a NEP model and can be used directly as an input to the gpumd executable for running atomistic simulations. To use the gpumd executable, one needs to prepare another input file, run.in, which specifies the simulation process. Several examples are presented in Sect. V below.

III Performance of NEP models

III.1 NEP3 vs NEP2

In this section, we demonstrate the workflow of using the nep executable to train NEPs and show the enhanced accuracy of NEP3 compared to NEP2 due to the improved radial functions. Here we use the MgAlCu alloy system, which has been studied previously using a DP [122]. There are 141,409 structures and we used 90% of this set for training and 10% for testing. There are tools in the gpumd package for preparing the required train.in and test.in input files for the nep executable.

Table 1: RMSEs for energies, forces, and virials for the MgAlCu alloy system using NEP2 and NEP3.
RMSE NEP2 NEP3
Energies (meV/atom\mathrm{meV}\text{/}\mathrm{atom}) 11 10
Forces (meV/Å\mathrm{meV}\text{/}\mathrm{\text{Å}}) 68 63
Virials (meV/atom\mathrm{meV}\text{/}\mathrm{atom}) 43 41

The other input file needed for the nep executable is nep.in. For our test using NEP2, this file reads as follows:

version 2type 3 Cu Al Mgcutoff 6 3.5n_max 15 10l_max 4neuron 50lambda_1 0.05lambda_2 0.05lambda_e 1.0lambda_f 1.0lambda_v 0.1batch 1000population 50generation 500000For our test using NEP3, it reads:

version 3type 3 Cu Al Mgcutoff 6 3.5n_max 12 8basis_size 12 8l_max 4neuron 50lambda_1 0.05lambda_2 0.05lambda_e 1.0lambda_f 1.0lambda_v 0.1batch 1000population 50generation 500000

The only differences between the hyperparameters used for NEP2 and NEP3 are related to the radial functions. For NEP2, we use nmaxR=15n_{\mathrm{max}}^{\mathrm{R}}=15 and nmaxA=10n_{\mathrm{max}}^{\mathrm{A}}=10. For NEP3, we use nmaxR=NbasR=12n_{\mathrm{max}}^{\mathrm{R}}=N_{\mathrm{bas}}^{\mathrm{R}}=12 and nmaxA=NbasA=8n_{\mathrm{max}}^{\mathrm{A}}=N_{\mathrm{bas}}^{\mathrm{A}}=8. The reason for using smaller nmaxRn_{\mathrm{max}}^{\mathrm{R}} and nmaxAn_{\mathrm{max}}^{\mathrm{A}} values in NEP3 compared to NEP2 is that one radial function in NEP3 is a linear combination of a few basis functions, while one radial function in NEP2 is simply one basis function. Even in this case, NEP3 can achieve a noticeably higher accuracy compared to NEP2, as shown in Table 1. With these parameters, NEP3 still has a similar computational speed as NEP2 in MD simulations, reaching about 1.5×107 atom step/s1.5\text{\times}{10}^{7}\text{\,}\mathrm{atom}\text{\,}\mathrm{step}\text{/}\mathrm{s} using one Tesla A100 GPU.

The nep executable produces a number of output files that can be used to examine the training/testing results in detail. Figure 2 shows the training and testing RMSEs as well as the loss function related to the regularization (from the loss.out file). The training RMSEs exhibit oscillations because of the use of mini-batches (with a batch size of 1,000). The test RMSEs closely follow the training RMSEs, which indicates the very good interpolation capability of NEP. In other words, there is no sign of over-fitting. As shown in Ref. 41, a proper regularization is crucial to prevent possible over-fitting in NEP models. Figure 3 shows the results from the force_test.out file. We can see that both NEP2 and NEP3 achieve a rather high accuracy here.

Figure 2: Evolution of the various terms in the loss function (41) with respect to the generation in the SNES [68] training algorithm for the MgAlCu training and test data sets. [122]
Refer to caption
Figure 3: Forces from NEP2 and NEP3 models against the target DFT values for the MgAlCu test set. The solid line represents the identity function that serves as a guide to the eye.

III.2 Comparison of NEP3 with other MLPs

III.2.1 A general-purpose silicon data set

We use the general-purpose silicon training data set from Ref. 123 to test convergence with respect to some hyperparameters and compare the results from an implementation of the ACE [48]. This data set consists of 2,475 structures, including bulk crystal structures, sp2 bonded structures, dimers, liquid structures, amorphous structures, diamond structures with surfaces or vacancies, and several other defective structures. Every structure has an energy, but not all the structures have virial data. For details on the reference DFT calculations, the reader is referred to Ref. 123.

We use the same cutoff distance of 5 Å5\text{\,}\mathrm{\text{Å}} for the radial and angular parts and set nmaxR=nmaxA=NbasR=NbasA=10n_{\mathrm{max}}^{\mathrm{R}}=n_{\mathrm{max}}^{\mathrm{A}}=N_{\mathrm{bas}}^{\mathrm{R}}=N_{\mathrm{bas}}^{\mathrm{A}}=10. We consider using 3-body descriptor components only (lmax3​b=4l_{\mathrm{max}}^{\mathrm{3b}}=4, lmax4​b=0l_{\mathrm{max}}^{\mathrm{4b}}=0, lmax5​b=0l_{\mathrm{max}}^{\mathrm{5b}}=0), using both 3-body and 4-body descriptor components (lmax3​b=4l_{\mathrm{max}}^{\mathrm{3b}}=4, lmax4​b=2l_{\mathrm{max}}^{\mathrm{4b}}=2, lmax5​b=0l_{\mathrm{max}}^{\mathrm{5b}}=0), and using up to 5-body descriptor components (lmax3​b=4l_{\mathrm{max}}^{\mathrm{3b}}=4, lmax4​b=2l_{\mathrm{max}}^{\mathrm{4b}}=2, lmax5​b=1l_{\mathrm{max}}^{\mathrm{5b}}=1). Other common hyperparameters are as follows: λ1=λ2=0.05\lambda_{1}=\lambda_{2}=0.05, λe=1\lambda_{\mathrm{e}}=1, λf=1\lambda_{\mathrm{f}}=1, λv=0.1\lambda_{\mathrm{v}}=0.1, Nneu=50N_{\mathrm{neu}}=50, Nbat=fullN_{\mathrm{bat}}=\rm{full}, Npop=50N_{\mathrm{pop}}=50, and Ngen=3×105N_{\mathrm{gen}}=3\times 10^{5}.

Figure 4: Force RMSE against computational cost for NEP models for silicon compared to an implementation of the ACE approach (the faster recursive approach as in Ref. 48). A serial C++ implementation of the NEP approach has been tested using an Intel i7-8750H CPU @ 2.2 GHz. For comparison, the ACE potential was implemented in Julia and tested using an Intel i7-7820HQ CPU @ 2.9 GHz. [48]

Figure 4 shows the force RMSE versus the computational cost of force evaluation for NEP models and previous results for an implementation [48] of the ACE potential [46] based on linear regression. While the ACE potential shows a strong dependence of the accuracy on the maximum correlation order of the angular descriptor components, our NEP shows a much weaker dependence, although considering 4-body and 5-body correlations indeed helps to increase the accuracy to some degree. With 3-body descriptor components only, the ACE potential has a force RMSE above 1 eV/Å1\text{\,}\mathrm{eV}\text{/}\mathrm{\text{Å}}, while the 3-body NEP model already achieves an accuracy of about 0.1 eV/Å0.1\text{\,}\mathrm{eV}\text{/}\mathrm{\text{Å}}. To achieve the same accuracy as the NEP models, one needs to consider up to 6-body descriptor components in the ACE approach and the computational cost is a few times larger than that of the NEP models. The reason for the relative lower cost of the NEP models compared to the ACE model is probably due to the use of a neural network as the regression method instead of linear regression as used in the ACE approach. Using linear regression, the descriptor needs to be rather complete, which can easily lead to more than 10410^{4} descriptor components [48], while the descriptor vector lengths range from 55 to 77 in the present NEP models. With a reduced descriptor length, the completeness [48] of the descriptor will be reduced to some degree, but the incompleteness of the descriptor can be (partially) compensated by the neural network and the overall computational cost at a given target accuracy can be lower than using a large number of descriptor components and linear regression. Indeed, within the framework of NN-body iterative contraction of equivariants (NICE), nonlinear neural network regression has been shown to be able to achieve higher accuracy than linear regression at least in the limit of large training set size, [124] using a descriptor up to the same level of the NN-body correlation.

III.2.2 Azobenzene molecule

Our next example is the largest molecule, azobenzene, in the MD17 data set [125]. This data set has been revised later [126] to ensure more strict convergence in the DFT calculations, which is referred to as the revised MD17 data set (rMD17). Here, we use the first train-test split as reported in rMD17 [126], with 1,000 training structures and 1,000 testing structures randomly chosen from a MD trajectory at 500 K500\text{\,}\mathrm{K}. There are thus 1,000 target energies and 72,000 target force components in the training data set but there are no target virials.

As we have confirmed that adding 4-body and 5-body descriptor components can lead to higher accuracy, here we use lmax3​b=4l^{\mathrm{3b}}_{\mathrm{max}}=4, lmax4​b=2l^{\mathrm{4b}}_{\mathrm{max}}=2, lmax5​b=1l^{\mathrm{5b}}_{\mathrm{max}}=1 and consider different values of the radial function hyperparameters. For simplicity, we set nmaxR=NbasRn^{\mathrm{R}}_{\mathrm{max}}=N^{\mathrm{R}}_{\mathrm{bas}} and nmaxA=NbasAn^{\mathrm{A}}_{\mathrm{max}}=N^{\mathrm{A}}_{\mathrm{bas}} and consider the following combinations of parameters: nmaxR=nmaxA=(6,4)n^{\mathrm{R}}_{\mathrm{max}}=n^{\mathrm{A}}_{\mathrm{max}}=(6,4), (9,6)(9,6), (12,8)(12,8), and (15,10)(15,10). The other hyperparameters are: rcR=6r_{\mathrm{c}}^{\mathrm{R}}=6 Å, rcA=4r_{\mathrm{c}}^{\mathrm{A}}=4 Å, λ1=λ2=0.02\lambda_{1}=\lambda_{2}=0.02, λe=1\lambda_{\mathrm{e}}=1, λf=1\lambda_{\mathrm{f}}=1, λv=0\lambda_{\mathrm{v}}=0, Nneu=50N_{\mathrm{neu}}=50, Nbat=N_{\mathrm{bat}}= full, Npop=50N_{\mathrm{pop}}=50, and Ngen=106N_{\mathrm{gen}}=10^{6}.

Figure 5 shows the force mean absolute error (MAE) versus the computational cost of force evaluation for our NEP and some other MLPs as reported in Ref. 47, including ANI [15], GAP [8], sGDML [24], and a linear-regression based ACE potential [47] (similar to but not identical to the one in Ref. 48). With increasing nmaxRn_{\mathrm{max}}^{\mathrm{R}} and nmaxAn_{\mathrm{max}}^{\mathrm{A}}, both the accuracy and computational cost increase quickly. With nmaxR=15n_{\mathrm{max}}^{\mathrm{R}}=15 and nmaxA=10n_{\mathrm{max}}^{\mathrm{A}}=10, the NEP approach achieves an accuracy between ANI and GAP, but it is more than one order of magnitude faster than ANI and more than two orders of magnitude faster than GAP. At this level of accuracy, the NEP models are also several times faster than linear-ACE potentials, similar to the case of silicon above (Fig. 4). Similar to the case of of silicon, we propose that the superior cost effectiveness of the NEP models compared to the linear-ACE potentials is due to the much smaller descriptor vector size in NEP, which ranges from 32 to 82 in the NEP models, but from 1,700 to 122,000 in the linear-ACE potentials for the test cases in Fig. 5.

Figure 5: Force MAE against computational cost from NEP models and several other MLPs for azobenzene as reported in Ref. 47. To be consistent with Ref. 47, we tested our serial C++ code using an Intel Xeon Gold 5218 @ 2.3 GHz.

III.2.3 Carbon data set

In this section, we use the training and test data sets for the carbon system with various phases [10] to compare the NEP approach with other MLPs in terms of accuracy, speed, and memory usage. The MLPs to be compared include DP [17], GAP [8], MTP [30], and REANN [45]. The training data set comprises 4,080 structures in total, including bulk crystal structures, bulk liquid and amorphous structures, amorphous surfaces, and isolated dimer structures. The testing data set comprises 450 structures similar to those in the training data set, but excluding the dimer ones. There are 256,628 and 28,337 atoms in the training and testing data sets, respectively. Each structure has one target energy and each atom has three target force components. Some structures also have target virials. For more details on the data sets, see Ref. 10.

Table 2: Performance comparison between NEP models and other MLPs for the carbon test set from Ref. 10. Energy RMSE Δ​E\Delta E and virial RMSE Δ​W\Delta W are in units of meV/atom\mathrm{meV}\text{/}\mathrm{atom} while the force RMSE Δ​F\Delta F is in units of meV/Å\mathrm{meV}\text{/}\mathrm{\text{Å}}. Computational speed is measured in atom step/ms\mathrm{atom}\text{\,}\mathrm{step}\text{/}\mathrm{ms}. NmaxN_{\mathrm{max}} is the maximum number of atoms that can be simulated using one Tesla V100 GPU for the three GPU-accelerated codes. For GAP [8, 10] and MTP [30, 31], 72 Intel Xeon-Gold 62406240 CPU cores were used. For DP [17] after compression, NEP and REANN [45, 32], a 32-GB Tesla V100 GPU was used.
MLP Δ​E\Delta E Δ​F\Delta F Δ​W\Delta W Speed NmaxN_{\mathrm{max}}
GAP 4646 1,1001,100 NA 6.1 NA
DP (se2) 8080 1,1001,100 250 290 240×103240\times 10^{3}
DP (se2+se3) 4444 800800 170 150 220×103220\times 10^{3}
MTP (44 Å) 3636 650650 180 110 NA
MTP (55 Å) 3535 630630 200 61 NA
MTP (66 Å) 3535 650650 220 27 NA
NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) 4242 690690 160 3,600 4,100×1034,100\times 10^{3}
NEP (3.7 Å3.7\text{\,}\mathrm{\text{Å}}) 4444 700700 170 4,600 5,800×1035,800\times 10^{3}
REANN (3 Å3\text{\,}\mathrm{\text{Å}}) 4141 700700 NA 280 290×103290\times 10^{3}
REANN (4 Å4\text{\,}\mathrm{\text{Å}}) 3131 640640 NA 170 180×103180\times 10^{3}
REANN (6 Å6\text{\,}\mathrm{\text{Å}}) 2828 670670 NA 62 64×10364\times 10^{3}

In the case of the NEP approach, we used the NEP3 form and considered two sets of hyperparameters. In the first one, we set rcR=4.2 År_{\mathrm{c}}^{\mathrm{R}}=$4.2\text{\,}\mathrm{\text{Å}}$, rcA=3.7 År_{\mathrm{c}}^{\mathrm{A}}=$3.7\text{\,}\mathrm{\text{Å}}$, nmaxR=NbasR=10n_{\mathrm{max}}^{\mathrm{R}}=N_{\mathrm{bas}}^{\mathrm{R}}=10, nmaxA=NbasA=8n_{\mathrm{max}}^{\mathrm{A}}=N_{\mathrm{bas}}^{\mathrm{A}}=8, lmax3​b=4l^{\mathrm{3b}}_{\mathrm{max}}=4, lmax4​b=2l^{\mathrm{4b}}_{\mathrm{max}}=2, lmax5​b=1l^{\mathrm{5b}}_{\mathrm{max}}=1, λ1=λ2=0.05\lambda_{1}=\lambda_{2}=0.05, λe=1\lambda_{\mathrm{e}}=1, λf=1\lambda_{\mathrm{f}}=1, λv=0.1\lambda_{\mathrm{v}}=0.1, Nneu=100N_{\mathrm{neu}}=100, Nbat=fullN_{\mathrm{bat}}=\rm{full}, Npop=50N_{\mathrm{pop}}=50, and Ngen=5×105N_{\mathrm{gen}}=5\times 10^{5}. This model is labelled “NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}})” in Table 2. In the second one, we make the following changes compared to the first one: rcR=3.7 År_{\mathrm{c}}^{\mathrm{R}}=$3.7\text{\,}\mathrm{\text{Å}}$, rcA=3.2 År_{\mathrm{c}}^{\mathrm{A}}=$3.2\text{\,}\mathrm{\text{Å}}$, lmax5​b=0l^{\mathrm{5b}}_{\mathrm{max}}=0, and Nneu=50N_{\mathrm{neu}}=50. This model is labelled “NEP (3.7 Å3.7\text{\,}\mathrm{\text{Å}})” in Table 2.

For DP, we used the DeePMD-kit package [17] and the smooth edition [19]. We trained two versions of DP, one using the se_a descriptor (with a cutoff of 6 Å6\text{\,}\mathrm{\text{Å}}) only, and the other using a combination of se_e2_a (with a cutoff of 6 Å6\text{\,}\mathrm{\text{Å}} and se_e3 (with a cutoff of 3.8 Å3.8\text{\,}\mathrm{\text{Å}}). These two versions are labelled “DP (se2)” and “DP (se2+se3)” in Table 2, respectively. The size of the embedding net is (25,50,100)(25,50,100) for the se_a and se_e2_a descriptors and (20,40,80)(20,40,80) for the se_e3 descriptor, and the size of the fitting net is (240,240,240)(240,240,240). The learning rate decreases exponentially from 10−310^{-3} to 10−810^{-8}. The weighting parameters for energy, force, and virial have a starting value of 0.02, 1,000, and 0.01, respectively, which are linearly changed to 1, 1, and 0.1 during the training process. The number of training steps is 10710^{7}, which is sufficiently large.

For GAP [8, 10], we directly took the results from Ref. 10. The nmaxn_{\mathrm{max}} and lmaxl_{\mathrm{max}} for the smooth overlap of atomic positions (SOAP) descriptor were both set to 88 [10]. There were also separate low-dimensional 2-body and 3-body components in this GAP [10].

For MTP [30], we used the mlip package [31]. The descriptor level of the MTP is set to 2222. We considered three cutoff distances: 4 Å4\text{\,}\mathrm{\text{Å}}, 5 Å5\text{\,}\mathrm{\text{Å}}, and 6 Å6\text{\,}\mathrm{\text{Å}}, labelled “MTP (4 Å4\text{\,}\mathrm{\text{Å}})”, “MTP (5 Å5\text{\,}\mathrm{\text{Å}})”, and “MTP (6 Å6\text{\,}\mathrm{\text{Å}})”, respectively, in Table 2.

For REANN [45], we used the reann package [32]. We considered three cutoff distances: 3 Å3\text{\,}\mathrm{\text{Å}}, 4 Å4\text{\,}\mathrm{\text{Å}}, and 6 Å6\text{\,}\mathrm{\text{Å}}, labelled “REANN (3 Å3\text{\,}\mathrm{\text{Å}})”, “REANN (4 Å4\text{\,}\mathrm{\text{Å}})”, and “REANN (6 Å6\text{\,}\mathrm{\text{Å}})”, respectively, in Table 2. The weighting parameter for energy is kept at 1 and that for force is decreased from 10 to 0.5 during the training process. A batch-size of 32 is used (we have tried to use a larger batch-size and it turned out to exceed the memory limit of a 32-GB V100). The sizes of the neural network for the atom energy and the orbital coefficients are both (64,64)(64,64). The learning rate decreases exponentially from 10−310^{-3} to 10−710^{-7}. The number of training epochs is 10410^{4}.

For all MLPs, we list the RMSEs for energies, forces, and virials (calculated from 6 independent components) in Table 2. The MTP and REANN models show the best accuracy in energies and forces, and GAP and DP the worst. The accuracy of the NEP models is close to those of MTP and REANN. The virial RMSE is missing for GAP as no predicted virial data have been provided [10]. It is also missing for REANN because the virial has not been formulated in this MLP. For the MLPs with available virial data, the NEP models achieve the highest accuracy. Therefore, we can say that the NEP models at least have an above-average accuracy for this carbon data set.

Figure 6: Force RMSE against computational cost from NEP models and other MLPs for the carbon data set from Ref. 10.

With the accuracy comparison results in mind, we next compare the computational performance in realistic atomistic simulations. Here, we run MD simulations for a cubic cell of diamond in the isothermal ensemble at 300 K300\text{\,}\mathrm{K} for 100 steps and output some basic thermodynamic properties every 10 steps. Based on the MD simulations, we measure the computational speed as the product of the number of atoms and the number of steps divided by the total wall time used. Three of the MLPs (DP, REANN, NEP) have been implemented on GPUs and we thus use an Nvidia Tesla V100 GPU for the test. For the other two MLPs for which there are only CPU versions available (GAP and MTP) we use 72 Intel Xeon-Gold 6240 cores. The CPU and GPU resources might have unequal financial costs, but one can make suitable conversions of the results presented here to other computational environments. For the CPU-based MLPs, MTP shows much higher computational speed than GAP, which is consistent with previous tests [30]. For the GPU-based MLPs, DP (after model compression) and REANN have comparable speed, while the NEP models are more than one order of magnitude faster. Figure 6 shows that the NEP models substantially lower the Pareto front of accuracy-versus-cost that can be achieved by the other MLPs.

Interestingly, the GPU memory usage seems to be correlated to the computational speed: the maximum number of atoms NmaxN_{\mathrm{max}} that can be simulated using one Tesla V100 GPU is roughly proportional to the computational speed. This comparison highlights the superior computational performance of the NEP approach as implemented in gpumd in terms of both computational speed and memory efficiency, which is crucial for tackling challenging applications that require large-scale and long-time atomistic simulations.

IV Active learning based on the latent space

Apart from regression accuracy and MD speed, training data preparation is an important aspect of MLPs. Because quantum-mechanical calculations are usually time consuming, it is desirable to construct a minimal training data set for a given application. One strategy for achieving this is to use active learning (AL), which uses query criteria to determine whether or not a new training sample should be included into an existing training set to improve the model accuracy and generalization capability. Many AL schemes have been proposed for MLPs, including the ensemble (or query-by-committee) method [127, 128, 129], the dropout method [130], methods based on feature-space distance measuring [131] and entropy maximization in the descriptor space [132], and methods based on optimal design [133, 134]. Here, we propose an AL scheme based on the latent space of a pre-trained NEP model. This AL scheme has been inspired by the work of Janet et al. [135], who have shown that distance in the latent space provides a good quantitative uncertainty metric to be used in an AL scheme.

There is no unique definition of the latent space. Here, we define it as an NneuN_{\mathrm{neu}}-dimensional space spanned by the vectors whose components are the product of the states of the hidden-layer neurons and the connection weights between them and the output layer. To compute the latent-space vector for a structure, one must train a NEP model first, but this can be achieved by using a small initial training data set. Then one can use the pre-trained NEP model to compute the latent-space vectors for many structures, either those in the training data set or new ones that have no target values (energy, forces, and virials) yet. This allows one to generate target values (via quantum-mechanical calculations) for a number of structures that have relatively large distances to existing points in the latent space. This procedure can be iterated by updating the training data set and the NEP model in alternating fashion. During this process, the existing NEP model can be used to create the new structures to be examined, using various sampling techniques in atomistic simulations.

Figure 7: Normalized explained variance ratio of the first 30 principal components calculated from the 4,080 structures in the original training data set [10] based on the initial NEP model trained using 200 structures.
Refer to caption
Figure 8: Distribution of the 4,080 carbon structures in the full training data set [10] in the 2D principal component (PC) space (spanned by PC 1 and PC 2) as reduced from the latent space that was constructed using the initial NEP model trained using 200 structures. (a) bulk amorphous/liquid structures, (b) crystals including sp2 graphite and sp3 diamond structures, (c) dimers, and (d) surface amorphous structures. The color bar represents the density of structures in the 2D PC space.

We take the carbon data set as a concrete example to illustrate the idea outlined above. To this end, suppose we only have 200 structures randomly selected out of the 4,080 ones in the original training data set. We first train an initial NEP using these 200 structures, adopting the same hyperparameters as used for the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model in Table 2. Assuming that we have obtained the remaining structures as in the original training data set by various means, we then compute the latent space vectors for all the 4,080 structures. One can define a distance in the high-dimensional latent space, but it turns out that the high dimension can be effectively reduced using principal component (PC) analysis. The explained variance ratios of the first 30 PCs are shown in Fig. 7. The first two leading PCs contribute more than 70% to the total dimensions, allowing us to visualize the distribution of structures in a 2D PC space, as shown in Fig. 8. It can be seen that both dimers and crystals (with or without defects) occupy a small area in the PC space. On the other hand, both bulk and surface amorphous/liquid structures occupy large areas that are almost overlapping. Furthermore, the high density part in Fig. 8a indicates that there are relatively more bulk amorphous and liquid structures in the full training data set.

Refer to caption
Figure 9: (a) Energy, (b) force, and (c) virial values from the NEP models for carbon constructed using the full training data set (4,080 structures, labelled “Full”) and the training data set constructed based on active learning (786 structures, labelled “Active”), in comparison to the DFT reference data for the test data set (450 structures).[10]

Based on these observations, we construct a new training data set which includes all the dimers (30 in total) and crystals (356 in total), and 400 bulk amorphous/liquid structures. Using these 786 structures, we train a new NEP model using the same hyperparameters as for the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model in subsubsection III.2.3). The energy, force, and virial RMSEs from this NEP model are 45 meV/atom45\text{\,}\mathrm{meV}\text{/}\mathrm{atom}, 700 meV/Å700\text{\,}\mathrm{meV}\text{/}\mathrm{\text{Å}}, and 190 meV/atom190\text{\,}\mathrm{meV}\text{/}\mathrm{atom}, respectively, for the same test data set as used in subsubsection III.2.3, which are very close to those for the NEP model trained using the full training data set. Figure 9 shows that the AL-based NEP model indeed performs very well in the various predicted values. The force RMSE (660 meV/Å660\text{\,}\mathrm{meV}\text{/}\mathrm{\text{Å}}) for the AL-based NEP model is only slightly higher than that for the NEP model trained using the full training data set (650 meV/Å650\text{\,}\mathrm{meV}\text{/}\mathrm{\text{Å}}). This is a notable result since we have not included a single surface amorphous structure into the training data set for the AL-based NEP model. This shows that the distance in the latent space (and the reduced PC space) indeed provides a reliable metric for selecting new samples for the construction of accurate and transferable MLPs. The present results also indicate that the NEP approach is quite data efficient, which we attribute to the relatively simple neural-network model and the inclusion of regularization terms in the loss function. As a further demonstration of the reliability of the AL-based NEP model, we show in Sect. V below that it performs equally well as the NEP model trained against the full training data set in an MD simulation covering a large range of temperatures.

V Examples for applications of NEP models

In this section, we demonstrate the application of NEP models in atomistic simulations. To this end, we employ the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model from Table 2, if not stated otherwise.

V.1 Lattice constant

We begin with a simple static calculation and determine the zero-temperature lattice constant of diamond by calculating a cohesive energy curve. The run.in input file reads:

potential potentials/nep/C_2022_NEP3.txt 0compute_cohesive 0.98 1.03 51The potential keyword specifies the NEP model to be used and the compure_cohesive keyword is used to invoke the cohesive energy calculation. The lattice constant thus obtained is 3.530 Å3.530\text{\,}\mathrm{\text{Å}}, which is very close to the DFT reference value obtained using the local density approximation (LDA), see Table 3.

Table 3: Structural and elastic properties of diamond from the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model in comparison to GAP and DFT-LDA results. [10]
DFT-LDA GAP NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}})
aa (Å\mathrm{\text{Å}}) 3.5323.532 3.5393.539 3.5303.530
C11C_{11} (GPa\mathrm{GPa}) 1,1011,101 1,0901,090 1,1341,134
C12C_{12} (GPa\mathrm{GPa}) 148148 112112 153153
C44C_{44} (GPa\mathrm{GPa}) 592592 594594 605605

V.2 Elastic constants

Next we compute the zero-temperature elastic constants, for which the run.in input file reads:

potential potentials/nep/C_2022_NEP3.txt 0compute_elastic 0.01 cubicHere, the compute_elastic keyword is used to initiate the calculation of the three independent elastic constant components (C11C_{11}, C12C_{12}, and C44C_{44}) using the energy-strain relation with ±1%\pm 1\% strain values. The computed elastic constants are presented in Table 3. The elastic constants from the NEP model are within 4% of the reference DFT-LDA values. For comparison, the C12C_{12} value from GAP is about 24% smaller than the reference DFT-LDA value. We note that including virial information during training is crucial for obtaining accurate elastic properties, as can be seen from Fig. 10. In other words, fitting to energy and force data alone does not guarantee an accurate description of virials. Since the calculation of the heat current involves virial terms, see Eq. (33), this is also important for heat transport applications, as has already been pointed out by Shimamura et al. [136].

Refer to caption
Figure 10: Virial values from the NEP models in comparison with DFT-LDA data for the carbon data set. [10]

V.3 Tensile loading of diamond

All the calculations above are static ones at zero temperature. Here, we use MD simulations to study the fracture of diamond under uniaxial tensile loading. The run.in input file reads:

potential potentials/nep/C_2022_NEP3.txt 0velocity 300ensemble npt_ber 300 300 100 0 0 0 1000 1000 1000 1000time_step 1run 100000ensemble npt_scr 300 300 100 0 0 0 1000 1000 1000 1000deform 1.42e-5 0 0 1dump_thermo 100dump_position 10000run 5000000

The simulated model is a periodic supercell comprising of 5×5×405\times 5\times 40 conventional cubic unit cells and hence 8,000 atoms. We first equilibrate the system at 300 K300\text{\,}\mathrm{K} and zero pressure using the NPT ensemble via the Berendsen thermostat and barostat [75] for 100 ps100\text{\,}\mathrm{ps}. Then we switch to the Bussi-Donadio-Parrinello thermostat [79] and the Bernetti-Bussi barostat [82] in the production stage, deforming the simulation box in the zz direction with a strain rate of 1×108 /s1\text{\times}{10}^{8}\text{\,}\text{/}\mathrm{s} for 5 ns5\text{\,}\mathrm{ns} up to a strain of 50%, while the pressures in the xx and yy directions are set to zero. Based on the output thermodynamic quantities and trajectory, we can obtain the stress-strain relation as shown in Fig. 11 and identify a snapshot of the fracture process as shown in Fig. 12. The fracture is brittle with a fracture strength of about 200 GPa200\text{\,}\mathrm{GPa} at a strain of about 29%. We note that more independent simulations are needed to obtain statistically meaningful results beyond the current demonstration.

Figure 11: Stress-strain relation from a uniaxial tensile loading simulation of diamond using the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model for carbon.
Refer to caption
Figure 12: A snapshot sampled during the fracture process under uniaxial tensile loading.

V.4 Quenching

The carbon data set [10] is particularly suitable for studying liquid and amorphous carbon [10, 137, 138]. In this example, we use a melt-quench-anneal protocol similar to that used in Ref. 10 (but with ten times longer simulation time for the relaxation at each temperature and an extra relaxation at 1000 K) to generate amorphous carbon. The run.in file reads:

potential potentials/nep/C_2022_NEP3.txt 0velocity 9000ensemble nvt_lan 9000 9000 100time_step 1dump_thermo 10dump_position 1000run 30000ensemble nvt_lan 5000 5000 100dump_thermo 10dump_position 1000run 30000ensemble nvt_lan 5000 1000 100dump_thermo 10dump_position 100run 500ensemble nvt_lan 1000 1000 100dump_thermo 10dump_position 1000run 30000ensemble nvt_lan 300 300 100dump_thermo 10dump_position 1000run 30000

The initial simulation model is a cubic diamond supercell containing 64,000 atoms with a mass density of 3.0 g/cm33.0\text{\,}\mathrm{g}\text{/}{\mathrm{cm}}^{3}. The system is first quickly melted at 9.000 K9.000\text{\,}\mathrm{K} and then relaxed at 5.000 K5.000\text{\,}\mathrm{K}, followed by a quick quenching from 5.000 K5.000\text{\,}\mathrm{K} to 1.000 K1.000\text{\,}\mathrm{K} and further relaxation at 1.000 K1.000\text{\,}\mathrm{K} and 300 K300\text{\,}\mathrm{K}. Here, the Langevin thermostat [80] is used to control the temperature. The evolution of temperature and the ratio of sp3-bonded atoms as a function of simulation time are presented in Fig. 13a-b. The radial and angular distribution functions g⁡(r)g(r) and g⁡(θ)g(\theta) at 5.000 K5.000\text{\,}\mathrm{K} and 300 K300\text{\,}\mathrm{K} in Fig. 13(c)-(d) show that the system is in liquid and amorphous-solid states, respectively. We also performed the same MD simulation using the NEP model trained with the AL scheme in Sect. IV and we can see that it gives almost identical results as those from the NEP model trained using the full training data set. This further demonstrates the effectiveness of our AL scheme based on the latent space.

Figure 13: (a) Temperature and (b) ratio of sp3 bonded atoms as a function of simulation time as obtained using the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model. (c) Radial and (d) angular distribution functions at 5.000 K5.000\text{\,}\mathrm{K} and 300 K300\text{\,}\mathrm{K}.

V.5 Density of states and heat capacity of amorphous carbon

After obtaining amorphous carbon structures, we further study their thermal properties. In this example, we calculate the vibrational density of states (VDOS) for an amorphous carbon structure and then obtain the heat capacity with quantum corrections. The run.in file reads:

potential potentials/nep/C_2022_NEP3.txt 0velocity 300ensemble nvt_ber 300 300 100time_step 1run 10000ensemble nvecompute_dos 5 200 400run 100000

The VDOS ρ⁡(ω)\rho(\omega) is calculated from the velocity auto-correlation function (VACF) [139]. The VACF and VDOS are normalized to 3​N3N, where NN is the number of atoms. The per-atom heat capacity with quantum corrections at the temperature TT is then calculated as

C⁡(T)=1N​∫0∞d​ω2​π​ρ​(ω)​x2​ex(ex−1)2,C(T)=\frac{1}{N}\int_{0}^{\infty}\frac{d\omega}{2\pi}\rho(\omega)\frac{x^{2}e^{x}}{(e^{x}-1)^{2}}, (42)

where x=ℏ​ω/kB​Tx=\hbar\omega/k_{\mathrm{B}}T is the ratio of the vibrational energy ℏ​ω\hbar\omega and the thermal energy kB​Tk_{\mathrm{B}}T.

Figure 14: (a) VDOS of amorphous carbon with a density of 3.0 g/cm33.0\text{\,}\mathrm{g}\text{/}{\mathrm{cm}}^{3} at 300 K300\text{\,}\mathrm{K} as obtained using the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model. (b) Heat capacity with quantum corrections calculated from the VDOS via Eq. (42).

V.6 Thermal conductivity of amorphous carbon

In this last example, we calculate the thermal conductivity of our amorphous carbon sample using the HNEMD method and the related spectral decomposition method [87, 61]. The run.in file reads:

potential potentials/nep/C_2022_NEP3.txt 0velocity 300ensemble nvt_nhc 300 300 100time_step 1run 100000ensemble nvt_nhc 300 300 100compute_hnemd 1000 0 0 2e-4compute_shc 5 200 2 500 400run 2000000

Thermal conductivity calculations usually require a lot of data to reduce the statistical uncertainty. To this end, we perform a number of independent runs using the above inputs. In gpumd, the velocities are automatically initialized with different pseudo-random number seeds for different runs. Using the efficient HNEMD method, 5 independent runs (each with a production time of 2 ns2\text{\,}\mathrm{ns}) are sufficient to achieve high accuracy (small error bounds), as can be seen from Fig. 15a. The thermal conductivity of the amorphous carbon structure (with a mass density of 3.0 g/cm33.0\text{\,}\mathrm{g}\text{/}{\mathrm{cm}}^{3}) at 300 K300\text{\,}\mathrm{K} is determined to be 5.1±±0.1 W/(m K)5.1\pm\pm 0.1\text{\,}\mathrm{W}\text{/}\text{(}\mathrm{m}\text{\,}\mathrm{K}\text{)}, where the statistical error is calculated as the standard error [140]. The thermal conductivity calculated in this way is the classical value. For disordered materials, the thermal conductivity can be quantum corrected in a way similar to the quantum correction of the heat capacity [141]. To achieve this, we first calculate the classical spectral thermal conductivity [87, 61] κc​(ω)\kappa^{\mathrm{c}}(\omega) and include quantum corrections to obtain κq​(ω)\kappa^{\mathrm{q}}(\omega) as follows:

κq​(ω)=κc​(ω)​x2​ex(ex−1)2,\kappa^{\mathrm{q}}(\omega)=\kappa^{\mathrm{c}}(\omega)\frac{x^{2}e^{x}}{(e^{x}-1)^{2}}, (43)

where x=ℏ​ω/kB​Tx=\hbar\omega/k_{\mathrm{B}}T. Both κc​(ω)\kappa^{\mathrm{c}}(\omega) and κq​(ω)\kappa^{\mathrm{q}}(\omega) at 300 K300\text{\,}\mathrm{K} are shown in Fig. 15b. The quantum-corrected thermal conductivity at 300 K300\text{\,}\mathrm{K} is 3.2±±0.1 W/(m K)3.2\pm\pm 0.1\text{\,}\mathrm{W}\text{/}\text{(}\mathrm{m}\text{\,}\mathrm{K}\text{)}. A more systematic investigation of the thermal transport properties in disordered carbon systems will be presented elsewhere.

Figure 15: (a) Thermal conductivity of amorphous carbon as calculated from the HNEMD method using the NEP (4.2 Å4.2\text{\,}\mathrm{\text{Å}}) model. The thin solid lines are from 5 independent runs and the thick solid line is their average. The dashed lines represent the error bounds. (b) Classical and quantum-corrected spectral thermal conductivity as a function of the vibrational frequency.

VI Interface to other codes

VI.1 The gpyumd package

To help gpumd users generate input and process output files, we have developed a Python interface implemented in the gpyumd package. Reading, preparing, and writing xyz.in files is facilitated by the GpumdAtoms class. This class extends the Atoms class from the popular Atomic Simulation Environment (ase) Python package [142] to include gpumd-specific properties. A simple example of writing an xyz.in file is as follows:

from ase.lattice.cubic import Diamondfrom gpyumd.atoms import GpumdAtomsSi = GpumdAtoms(Diamond("Si",size=(10,10,10)))Si.set_max_neighbors(4)Si.set_cutoff(3)Si.write_gpumd()

The GpumdAtoms class also supports adding grouping methods, sorting atoms by group or type, generating basis.in and kpoints.in files for phonon calculations, and more.

The gpyumd package also has a Simulation class that can be used to generate valid run.in files. To do so, in addition to checking each keyword parameter, it verifies that group selections in each keyword, atom types in potential definitions, and atom ordering in the xyz.in file are consistent. As a simple demonstration, continuing from our previous code snippet, we can create a simple run.in file as follows:

import gpyumd.keyword as kwdfrom gpyumd.sim import Simulationsim = Simulation(Si)run = sim.add_run()keywords = [ kwd.Velocity(initial_temperature=300), kwd.TimeStep(dt_in_fs=1), kwd.Ensemble(ensemble_method=’nve’), kwd.DumpThermo(interval=1000), kwd.RunKeyword(number_of_steps=1e5)][run.add_keyword(x) for x in keywords]potential = kwd.Potential(filename=’Si.txt’, symbols=[’Si’])sim.add_potential(potential)sim.create_simulation()

The output files of gpumd can be read and processed using simple gpyumd functions such as load_thermo() for the thermo.out file, load_hac() for the hac.out file, etc. These functions return the data in convenient formats for data exploration in interactive environments such as Jupyter Notebooks.

VI.2 The calorine package

To provide a deeper integration of gpumd within a Python workflow, we also provide the Python package calorine. This section outlines the functionality of this package.

VI.2.1 ASE calculator

The calorine package extends the functionality of ase, implementing an ase Calculator class, which lets users calculate energies, forces, and stresses with a NEP model directly from Python. Under the hood, this calculator writes and reads the gpumd input and output files. A minimal script reads as follows:

from calorine.ase_calculator import GPUMD_NEPfrom ase.build import bulkcalculator = GPUMD_NEP(’nep.txt’)atoms = bulk(’Au’, a=4.1)atoms.set_calculator(calculator)e = atoms.get_potential_energy()

This approach can, for example, greatly simplify the calculation of a large number of pre-defined structures, and provides access to various complex structural relaxation schemes available in ase.

VI.2.2 Pybind11 interface

Calorine also interfaces directly to gpumd using a PyBind11 C++ interface. This enables easy access to the data structures associated with the NEP implementation in gpumd. At the moment the interface exposes a function for calculating the descriptors for an ase Atoms object, but this can be easily extended.

An example script for accessing the gpumd descriptors for a structure is given below.

from ase import Atomsfrom calorine.nepy.nep import get_descriptorsCO = Atoms( ’CO’, positions=[(0, 0, 0), (0, 0.0, 1.1)] )descriptors_CO = get_descriptors(CO)

VII Summary and conclusions

In summary, we have presented the various features of the open-source gpumd package, with a focus on recent developments that have enabled the generation and use of accurate and efficient NEP MLPs [41, 42]. Two improvements on the atomic-environment descriptor have been introduced: one is to change the radial functions from Chebyshev basis functions to linear combinations of the basis functions, and the other is to extend the angular descriptor components by considering some 4-body and 5-body contributions as in the ACE approach [46]. Both of these extensions are shown to improve the accuracy of NEP models further.

We have used a diverse set of materials to demonstrate that the NEP approach can achieve an above-average accuracy compared to many other state-of-the-art MLPs. In addition, NEP models can achieve a far superior computational efficiency: typical NEP models are more than one order of magnitude faster than other MLPs and similarly more memory efficient. The high efficiency of NEP models originates from many aspects, including a reasonably small descriptor dimension (usually smaller than 100) combined with a simple neural network model with a single hidden layer, carefully derived expressions of the descriptor components, a balanced choice of the radial and angular cutoff distances, and finally a carefully optimized GPU implementation. We present our algorithms in detail in Appendix B. The latent space in the simple neural network model of NEP models also allows us to construct an effective AL scheme that can be used to greatly reduce the computational efforts in the preparation of training data.

Apart from being highly efficient, gpumd is also user-friendly. It can both be used as a standalone package and be integrated with other packages such as ase [142] via the gpyumd and calorine Python packages. The use of an efficient derivative-free optimization algorithm (SNES) greatly simplifies the implementation and excludes the dependence of gpumd on any third-party machine-learning libraries, making the installation of gpumd effortless.

Finally, the NEP models trained using the nep executable can be directly used by the gpumd executable to perform atomistic simulations of various materials properties. To demonstrate the range of properties, length and time scales that can be accessed via this approach, we have presented a series of examples using a NEP model trained using a standard carbon data set [10].

One of the disadvantages of gpumd is that it is still not very feature-rich (as compared to similar packages such as lammps [71]). However, its open-source nature and the well-designed GPU-acceleration framework have been attracting more and more developers with diverse backgrounds who are enriching the features of gpumd at a fast pace.

Data availability

The training and testing results using the various MLPs as presented in Sect. III and Sect. IV are freely available via Zenodo [143]. The input and output files for the gpumd examples presented in Sect. V are included in the gpumd package (https://github.com/brucefan1983/GPUMD).

Code availability

The source code of gpumd and calorine is maintained via git and available at https://github.com/brucefan1983/GPUMD and https://gitlab.com/materials-modeling/calorine. The respective documentation pages can be found at https://gpumd.zheyongfan.org/ and calorine.materialsmodeling.org/. The latent-space-based active learning analyses have been facilitated by the PyNEP package available at https://github.com/bigd4/PyNEP. The source code and documentation of gpyumd will be published soon.

Acknowledgements.
Z.F. acknowledges support from the National Natural Science Foundation of China (NSFC) (No. 11974059). Y.W., K.S., J.L., and H.D. acknowledge the support from the National Key Research and Development Program of China (2018YFB0704300). The work at Nanjing University (J.J.W, Y.W., J.S.) is partially supported from the NSFC (grant nos. 12125404, 11974162, and 11834006), and the Fundamental Research Funds for the Central Universities. The calculations performed in Nanjing University were carried out using supercomputers at the High Performance Computing Center of Collaborative Innovation Center of Advanced Microstructures, the high-performance supercomputing center of Nanjing University. P.Y. and Z.Z. acknowledge support from the NSFC (No. 11932005). T.A-N. has been supported in part by the Academy of Finland through its QTF Centre of Excellence program (No. 312298) and Technology Industries of Finland Centennial Foundation Future Makers grant. Z.Z. and Y.C. are grateful for the research computing facilities offered by ITS, HKU. J.M.R., E.L., and P.E. acknowledge support from the Swedish Research Council (2018-06482, 2020-04935, 2021-05072) and the Swedish Foundation for Strategic Research (SSF) via the SwedNess program (GSn15-0008) as well as computational resources provided by the Swedish National Infrastructure for Computing (SNIC) at NSC, C3SE, and PDC partially funded by the Swedish Research Council (2018-05973).

Appendix A Some derivations on the heat current expressions

Here we show that the heat current expressions for the multi-body potentials as considered in Ref. 67 are special cases of the general expressions in our formulation. Without loss of generality, we take the 3-body potential as an example to show this.

Using the chain rule, we first rewrite Eq. (35) as

𝑱=−∑i∑j≠i𝒓i​j∂Ui∂𝒓j⋅𝒗j.\bm{J}=-\sum_{i}\sum_{j\neq i}\bm{r}_{ij}\frac{\partial U_{i}}{\partial\bm{r}_{j}}\cdot\bm{v}_{j}. (44)

Boone et al. [67] considered explicit mm-body potentials (m=2,3,4m=2,3,4) that are usually used in topological force fields for organics. For the 3-body potential considered therein, the site potential UiU_{i} of atom ii is taken as the average of the potentials of the triplets it belongs to (the first index denotes the central atom of a triplet):

Ui=13​∑k≠i∑l≠i(Ui​k​l+Uk​l​i+Ul​i​k).U_{i}=\frac{1}{3}\sum_{k\neq i}\sum_{l\neq i}\left(U_{ikl}+U_{kli}+U_{lik}\right). (45)

Substituting Eq. (45) into Eq. (44), we have

𝑱=−13∑i∑j≠i∑k≠i∑l≠i𝒓i​j∂(Ui​k​l+Uk​l​i+Ul​i​k)∂𝒓j⋅𝒗j.\bm{J}=-\frac{1}{3}\sum_{i}\sum_{j\neq i}\sum_{k\neq i}\sum_{l\neq i}\bm{r}_{ij}\frac{\partial\left(U_{ikl}+U_{kli}+U_{lik}\right)}{\partial\bm{r}_{j}}\cdot\bm{v}_{j}. (46)

Note that jj cannot be ii but could be kk or ll and we thus have

∂Ui​k​l∂𝒓j=δj​k​∂Ui​j​l∂𝒓j+δj​l​∂Ui​k​j∂𝒓j\frac{\partial U_{ikl}}{\partial\bm{r}_{j}}=\delta_{jk}\frac{\partial U_{ijl}}{\partial\bm{r}_{j}}+\delta_{jl}\frac{\partial U_{ikj}}{\partial\bm{r}_{j}} (47)

and similar expressions for ∂Uk​l​i∂𝒓j\frac{\partial U_{kli}}{\partial\bm{r}_{j}} and ∂Ul​i​k∂𝒓j\frac{\partial U_{lik}}{\partial\bm{r}_{j}}. Following Boone et al. [67] we can define

𝑭ji​j​l≡−∂Ui​j​l∂𝒓j.\bm{F}_{j}^{ijl}\equiv-\frac{\partial U_{ijl}}{\partial\bm{r}_{j}}. (48)

as the force acting on atom jj from the triplet i​j​lijl. Then we can write Eq. (46) as

𝑱\displaystyle\bm{J} =13​∑i∑j≠i∑k≠i𝒓i​j\displaystyle=\frac{1}{3}\sum_{i}\sum_{j\neq i}\sum_{k\neq i}\bm{r}_{ij}
(𝑭ji​j​k+𝑭ji​k​j+𝑭jj​k​i+𝑭jk​j​i+𝑭jj​i​k+𝑭jk​i​j)⋅𝒗j.\displaystyle\left(\bm{F}_{j}^{ijk}+\bm{F}_{j}^{ikj}+\bm{F}_{j}^{jki}+\bm{F}_{j}^{kji}+\bm{F}_{j}^{jik}+\bm{F}_{j}^{kij}\right)\cdot\bm{v}_{j}. (49)

By manipulating the dummy indices in the summation, it can be written as

𝑱\displaystyle\bm{J} =13​∑i∑j≠i∑k≠i(𝒓j​i+𝒓k​i)​𝑭ii​j​k⋅𝒗i\displaystyle=\frac{1}{3}\sum_{i}\sum_{j\neq i}\sum_{k\neq i}\left(\bm{r}_{ji}+\bm{r}_{ki}\right)\bm{F}_{i}^{ijk}\cdot\bm{v}_{i}
+(𝒓i​j+𝒓k​j)𝑭ji​j​k⋅𝒗j+(𝒓i​k+𝒓j​k)𝑭ki​j​k⋅𝒗k,\displaystyle+\left(\bm{r}_{ij}+\bm{r}_{kj}\right)\bm{F}_{j}^{ijk}\cdot\bm{v}_{j}+\left(\bm{r}_{ik}+\bm{r}_{jk}\right)\bm{F}_{k}^{ijk}\cdot\bm{v}_{k}, (50)

which corresponds to Eq. (18d) in Ref. 67.

Appendix B Algorithms

In this Appendix, we present the complete algorithms for evaluating the NEP energy, force, and virial expressions as implemented in gpumd. First we list all the relevant quantities:

  1. 1.

    NN is the total number of atoms.

  2. 2.

    NNiR\mathrm{NN}_{i}^{\mathrm{R}} is the number of neighbors of atom ii for the radial descriptor components.

  3. 3.

    NNiA\mathrm{NN}_{i}^{\mathrm{A}} is the number of neighbors of atom ii for the angular descriptor components.

  4. 4.

    NLi​mR\mathrm{NL}_{im}^{\mathrm{R}} is the index of the mm-th neighbor of atom ii for the radial descriptor components.

  5. 5.

    NLi​mA\mathrm{NL}_{im}^{\mathrm{A}} is the index of the mm-th neighbor of atom ii for the angular descriptor components.

  6. 6.

    {𝒓i}i=0N−1\{\bm{r}_{i}\}_{i=0}^{N-1} are the atom positions.

  7. 7.

    {Ui}i=0N−1\{U_{i}\}_{i=0}^{N-1} are the site energies.

  8. 8.

    {∂Ui/∂qνi}i=0N−1\{\partial U_{i}/\partial q^{i}_{\nu}\}_{i=0}^{N-1} are the derivatives of the site energies with respect to the descriptor components.

  9. 9.

    {∂Ui/∂𝒓i​j}i=0N−1\{\partial U_{i}/\partial\bm{r}_{ij}\}_{i=0}^{N-1} are the partial forces.

  10. 10.

    {𝑭i}i=0N−1\{\bm{F}_{i}\}_{i=0}^{N-1} are the forces on the atoms.

  11. 11.

    {𝐖i}i=0N−1\{\mathbf{W}_{i}\}_{i=0}^{N-1} are the virials on the atoms.

B.1 Application of the neural network

We use a simple feedforward neural network with a single hidden layer. We have done extensive tests and found that a single hidden layer is sufficient to achieve high accuracy for NEP models and using more hidden layers only reduces computational performance without a corresponding improvement in model accuracy. Using a single hidden layer, many variables can be defined as registers instead of global memory or local memory in the CUDA kernel, which are much more expensive to access. Therefore, using a single hidden layer can achieve a significantly higher computational performance in the parallelism scheme we adopted.

Listing 1: The function applying the feedforward neural network to the input descriptor vector to obtain the site energy of an atom and the derivative of the energy with respect to the descriptor components.
1 __device__ void apply_ann(
2 const int N_des,
3 const int N_neu,
4 const float* w0,
5 const float* b0,
6 const float* w1,
7 const float* b1,
8 const float* q,
9 float& energy,
10 float* energy_derivative)
11 {
12 for (int n = 0; n < N_neu; ++n) {
13 float w0_times_q = 0.0f;
14 for (int d = 0; d < N_des; ++d) {
15 w0_times_q += w0[n*N_des+d] * q[d];
16 }
17 float x1 = tanh(w0_times_q - b0[n]);
18 float tanh_der = 1.0f - x1 * x1;
19 energy += w1[n] * x1;
20 for (int d = 0; d < N_des; ++d) {
21 float y1 = tanh_der * w0[n*N_des+d];
22 energy_derivative[d] += w1[n] * y1;
23 }
24 }
25 energy -= b1[0];
26 }

The complete __device__ function applying the neural network is presented in 1. For the inputs, N_des is the dimension NdesN_{\mathrm{des}} of the descriptor vector, N_neu is the number of neurons NneuN_{\mathrm{neu}} in the hidden layer, w0 is the weight matrix 𝐰(0)\mathbf{w}^{(0)}, w1 is the weight vector 𝐰(1)\mathbf{w}^{(1)}, b0 is the bias vector 𝐛(0)\mathbf{b}^{(0)} in the hidden layer, b1 is the bias b(1)b^{(1)} in the output node, and q is the descriptor vector 𝐪i\mathbf{q}^{i}. For the outputs, energy is the site energy UiU_{i} and energy_derivative is the derivative of the site energy with respect to the descriptor components ∂Ui/∂qνi\partial U_{i}/\partial q_{\nu}^{i}. Note that we do not need to calculate the derivative of the energy (and other related quantities such as force and virial) with respect to the neural-network parameters, as required in the conventional gradient-descent approach. In our evolutionary algorithm approach, there is no need to calculate the derivative of the loss function with respect to any parameters. Therefore, our implementation is very simple regarding the neural network part and particularly, we do not make gpumd dependent on any third-party packages. This makes the installation of gpumd very simple and straightforward.

B.2 Energy and derivative of energy with respect to descriptor

In the first CUDA kernel (see algorithm 1), the thread associated with atom ii calculates the whole descriptor vector 𝐪i\mathbf{q}^{i} and calls the __device__ function in 1 to obtain the energy UiU_{i} and the derivatives ∂Ui/∂qνi\partial U_{i}/\partial q^{i}_{\nu}. The derivatives will be used in the next two CUDA kernels.

Algorithm 1 Pseudo code of the CUDA kernel for evaluating the descriptor vector qνiq^{i}_{\nu}, the per-atom energy UiU_{i}, and the derivatives of the energy with respect the descriptor components ∂Ui/∂qνi\partial U_{i}/\partial q^{i}_{\nu}.
1 Assign atom ii to CUDA thread ii
2 if i<Ni<N then
    3 Read position 𝒓i\bm{r}_{i} for atom ii from global memory
    4 for m=0m=0 to NNiR−1\mathrm{NN}_{i}^{\mathrm{R}}-1 do
       5 j←NLi​mRj\leftarrow\mathrm{NL}_{im}^{\mathrm{R}}
       6 Read in 𝒓j\bm{r}_{j} from global memory and calculate 𝒓i​j\bm{r}_{ij}
       7 𝒓i​j←\bm{r}_{ij}\leftarrow minimum image of 𝒓i​j\bm{r}_{ij}
       8 Calculate the radial functions gn​(ri​j)g_{n}(r_{ij}) according to Eq. (3)
       9 Accumulate the radial descriptor components according to Eq. (2)
    10 end for
    11 for n=0n=0 to nmaxAn_{\mathrm{max}}^{\mathrm{A}} do
       12 for m=0m=0 to NNiA−1\mathrm{NN}_{i}^{\mathrm{A}}-1 do
          13 j←NLi​mAj\leftarrow\mathrm{NL}_{im}^{\mathrm{A}}
          14 Read in 𝒓j\bm{r}_{j} from global memory and calculate 𝒓i​j\bm{r}_{ij}
          15 𝒓i​j←\bm{r}_{ij}\leftarrow minimum image of 𝒓i​j\bm{r}_{ij}
          16 Calculate the radial functions gn​(ri​j)g_{n}(r_{ij}) according to Eq. (3) but with rcRr_{\mathrm{c}}^{\mathrm{R}} changed to rcAr_{\mathrm{c}}^{\mathrm{A}}
          17 Accumulate the summations Sn,kS_{n,k} according to Eq. (20)
       18 end for
       19 Calculate the angular descriptor components for the current nn according to the equations in subsection II.4.
       20 Save the summations Sn,kS_{n,k} for the current nn to global memory for later use.
    21 end for
    22 Apply the neural network model to get the energy UiU_{i} and energy derivatives ∂Ui/∂qνi\partial U_{i}/\partial q^{i}_{\nu} from the descriptor qνiq^{i}_{\nu} and save them to global memory.
23 end if

B.3 Force and virial from the radial descriptor components

In the second CUDA kernel (see algorithm 2), the thread associated with atom ii first calculates the partial forces ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} and ∂Uj/∂𝒓j​i\partial U_{j}/\partial\bm{r}_{ji} related to the radial descriptor components, and then accumulates the force 𝑭i\bm{F}_{i} and virial 𝐖i\mathbf{W}_{i} on atom ii. For the radial descriptor components, ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} and ∂Uj/∂𝒓j​i\partial U_{j}/\partial\bm{r}_{ji} only differ a little and it is thus a good choice to calculate both within the CUDA kernel. This algorithm is very similar to that for EAM potentials [40] (which is an angular-independent many-body potential) as implemented in gpumd.

Algorithm 2 Pseudo code of the CUDA kernel for evaluating the force and virial from the radial descriptor components.
1 Assign atom ii to CUDA thread ii.
2 if i<Ni<N then
    3 Read position 𝒓i\bm{r}_{i} for atom ii from global memory
    4 for m=0m=0 to NNiR−1\mathrm{NN}_{i}^{\mathrm{R}}-1 do
       5 j←NLi​mRj\leftarrow\mathrm{NL}_{im}^{\mathrm{R}}.
       6 Read in 𝒓j\bm{r}_{j} from global memory and calculate 𝒓i​j\bm{r}_{ij}.
       7 𝒓i​j←\bm{r}_{ij}\leftarrow minimum image of 𝒓i​j\bm{r}_{ij}.
       8 Calculate the partial force ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} related to the radial descriptor components, i.e., the first term on the right hand side of Eq. (29).
       9 Similarly calculate the partial force ∂Uj/∂𝒓j​i\partial U_{j}/\partial\bm{r}_{ji}.
       10 Accumulate the force 𝑭i\bm{F}_{i} on atom ii according to Eq. (30) and Eq. (31).
       11 Accumulate the virial 𝐖i\mathbf{W}_{i} on atom ii according to Eq. Eq. 32.
    12 end for
13 end if

B.4 Partial forces from the angular descriptor components

Algorithm 3 Pseudo code of the CUDA kernel for evaluating the partial forces ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} for all atoms ii and all neighbors jj of ii from the angular descriptor components.
1 Assign atom ii to CUDA thread ii
2 if i<Ni<N then
    3 Read position 𝒓i\bm{r}_{i} for atom ii from global memory
    4 for m=0m=0 to NNiA−1\mathrm{NN}_{i}^{\mathrm{A}}-1 do
       5 j←NLi​mAj\leftarrow\mathrm{NL}_{im}^{\mathrm{A}}
       6 Read in 𝒓j\bm{r}_{j} from global memory and calculate 𝒓i​j\bm{r}_{ij}
       7 𝒓i​j←\bm{r}_{ij}\leftarrow minimum image of 𝒓i​j\bm{r}_{ij}
       8 Calculate the partial force ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} related to the angular descriptor components, i.e., the last three terms on the right hand side of Eq. (29).
       9 Save the partial force ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} to global memory for later use.
    10 end for
11 end if

In the third CUDA kernel (see algorithm 3), the thread associated with atom ii calculates the partial forces ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} related to the angular descriptor components and saves them to global memory, which is then used in the next CUDA kernel. For the angular descriptor components, ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} and ∂Uj/∂𝒓j​i\partial U_{j}/\partial\bm{r}_{ji} differ a lot and it is thus more efficient to use a two-kernel approach: using one CUDA kernel (the current one) to calculate the partial forces {∂Ui/∂𝒓i​j}\{\partial U_{i}/\partial\bm{r}_{ij}\} for all atom pairs (within the angular cutoff) and save them to global memory, and then using another CUDA kernel (the next one) to consume them.

B.5 Force and virial from the angular partial forces

Algorithm 4 Pseudo code of the CUDA kernel for evaluating the force and virial from the angular descriptor components.
1 Assign atom ii to CUDA thread ii.
2 if i<Ni<N then
    3 Read position 𝒓i\bm{r}_{i} for atom ii from global memory
    4 for m=0m=0 to NNiA−1\mathrm{NN}_{i}^{\mathrm{A}}-1 do
       5 j←NLi​mAj\leftarrow\mathrm{NL}_{im}^{\mathrm{A}}.
       6 Read in 𝒓j\bm{r}_{j} from global memory and calculate 𝒓i​j\bm{r}_{ij}.
       7 𝒓i​j←\bm{r}_{ij}\leftarrow minimum image of 𝒓i​j\bm{r}_{ij}.
       8 Read in the partial force ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} related to the angular descriptor components from global memory.
       9 Similarly read in Uj/∂𝒓j​iU_{j}/\partial\bm{r}_{ji} from global memory with some index manipulations.
       10 Accumulate the force 𝑭i\bm{F}_{i} on atom ii according to Eq. (30) and Eq. (31).
       11 Accumulate the virial 𝐖i\mathbf{W}_{i} on atom ii according to Eq. (32).
    12 end for
13 end if

After obtaining the partial force {∂Ui/∂𝒓i​j}\{\partial U_{i}/\partial\bm{r}_{ij}\} related to the angular descriptor components, we use a CUDA kernel (see algorithm 4) to accumulate the corresponding force and virial. In this kernel, we load the partial force ∂Ui/∂𝒓i​j\partial U_{i}/\partial\bm{r}_{ij} and ∂Uj/∂𝒓j​i\partial U_{j}/\partial\bm{r}_{ji}, which are related to each other by an exchange of atom indices ii and jj. Then we accumulate the force 𝑭i\bm{F}_{i} on atom ii according to Eqs. (30) and (31), and accumulate the virial 𝐖i\mathbf{W}_{i} on atom ii according to Eq. (32). This is a general CUDA kernel used in the gpumd package for all the angular-dependent many-body potentials, such as the Stillinger-Weber [73] and Tersoff potentials [38].

References